from __future__ import annotations

"""The webhook module contains the Webhook class, which is used to register a webhook to listen for incoming 
messages."""

import logging
import collections
import threading
from typing import TYPE_CHECKING, Callable, Any, Type
from pywa.types.base_update import BaseUpdate
from pywa import utils
from pywa.handlers import (
    Handler,  # noqa
    MessageHandler, CallbackButtonHandler, CallbackSelectionHandler, RawUpdateHandler, MessageStatusHandler
)
from pywa.utils import FlaskApp, FastAPIApp

if TYPE_CHECKING:
    from pywa import WhatsApp

__all__ = ["Webhook"]


class Webhook:
    """Register a webhook to listen for incoming messages."""

    def __init__(
            self,
            wa_client: WhatsApp,
            server: FlaskApp | FastAPIApp,
            verify_token: str,
            webhook_endpoint: str,
            filter_updates: bool
    ):
        self.handlers: dict[Type[Handler] | None, list[Callable[[WhatsApp, BaseUpdate | dict], Any]]] \
            = collections.defaultdict(list)
        self.wa_client = wa_client
        self.server = server
        self.verify_token = verify_token
        self.webhook_endpoint = webhook_endpoint
        self.filter_updates = filter_updates

        hub_vt = "hub.verify_token"
        hub_ch = "hub.challenge"

        if utils.is_flask_app(self.server):
            import flask

            @self.server.route(self.webhook_endpoint, methods=["GET"])
            def verify_token():
                if flask.request.args.get(hub_vt) == self.verify_token:
                    return flask.request.args.get(hub_ch), 200
                return "Error, invalid verification token", 403

            @self.server.route(self.webhook_endpoint, methods=["POST"])
            def webhook():
                threading.Thread(target=self.call_handlers, args=(flask.request.json,)).start()
                return "ok", 200

        elif utils.is_fastapi_app(self.server):
            import fastapi

            @self.server.get(self.webhook_endpoint)
            def verify_token(
                    token: str = fastapi.Query(..., alias=hub_vt),
                    challenge: str = fastapi.Query(..., alias=hub_ch)
            ):
                if token == self.verify_token:
                    return fastapi.Response(content=challenge, status_code=200)
                return fastapi.Response(content="Error, invalid verification token", status_code=403)

            @self.server.post(self.webhook_endpoint)
            def webhook(payload: dict = fastapi.Body(...)):
                threading.Thread(target=self.call_handlers, args=(payload,)).start()
                return fastapi.Response(content="ok", status_code=200)

        else:
            raise ValueError("The server must be a Flask or FastAPI app.")

    def call_handlers(self, update: dict) -> None:
        """Call the handlers for the given update."""
        handler = self._get_handler(update=update)
        for func in self.handlers[handler]:
            func(self.wa_client, handler.__constructor__(self.wa_client, update))
        for raw_update_func in self.handlers[RawUpdateHandler]:
            raw_update_func(self.wa_client, update)

    def _get_handler(self, update: dict) -> Type[Handler] | None:
        """Get the handler for the given update."""
        field = update["entry"][0]["changes"][0]["field"]
        value = update["entry"][0]["changes"][0]["value"]

        # The `messages` field needs to be handled differently because it can be a message, button, selection, or status
        if field == 'messages' and (
                not self.filter_updates or (value["metadata"]["phone_number_id"] == self.wa_client.phone_id)
        ):
            if 'messages' in value:
                if value["messages"][0]["type"] != "interactive":  # message
                    field = MessageHandler.__field_name__
                else:
                    if value["messages"][0]["interactive"]["type"] == "button_reply":  # button
                        field = CallbackButtonHandler.__field_name__
                    elif value["messages"][0]["interactive"]["type"] == "list_reply":  # selection
                        field = CallbackSelectionHandler.__field_name__

            elif 'statuses' in value:  # status
                field = MessageStatusHandler.__field_name__
            else:
                logging.warning(f"PyWa Webhook: Unknown message type: {value}")
        return next((h for h in Handler.__subclasses__() if h.__field_name__ == field), None)



"""Usefully filters to use in your handlers."""
from __future__ import annotations

__all__ = [
    "all_",
    "any_",
    "not_",
    "forwarded",
    "forwarded_many_times",
    "reply",
    "from_users",
    "from_countries",
    "text",
    "media",
    "image",
    "video",
    "audio",
    "document",
    "sticker",
    "reaction",
    "unsupported",
    "location",
    "contacts",
    "order",
    "callback",
    "message_status",
    "template_status",
]

import re
from abc import ABC, abstractmethod
from typing import Callable, TYPE_CHECKING, Iterable, TypeVar, TypeAlias, Type
from pywa import utils
from pywa.errors import WhatsAppError, ReEngagementMessage
from pywa.types import MessageType as Mt, Message as Msg, MessageStatus as Ms, MessageStatusType as Mst, \
    CallbackButton, CallbackSelection, TemplateStatus as Ts
from pywa.types.base_update import BaseUpdate

if TYPE_CHECKING:
    from pywa import WhatsApp as Wa

    MessageFilterT: TypeAlias = Callable[[Wa, Msg], bool]
    CallbackFilterT: TypeAlias = Callable[[Wa, CallbackButton | CallbackSelection], bool]
    MessageStatusFilterT: TypeAlias = Callable[[Wa, Ms], bool]
    TemplateStatusFilterT: TypeAlias = Callable[[Wa, Ts], bool]

T = TypeVar("T", bound=BaseUpdate)

forwarded: MessageFilterT = lambda _, m: m.forwarded  # Filter for forwarded messages.
"""
Filter for forwarded messages.

>>> filters.forwarded
"""

forwarded_many_times: MessageFilterT = lambda _, m: m.forwarded_many_times
"""
Filter for messages that have been forwarded many times.

>>> filters.forwarded_many_times
"""

reply: MessageFilterT = lambda _, m: m.reply_to_message is not None
"""
Filter for messages that reply to another message.

>>> filters.reply
"""


def all_(*filters: Callable[[Wa, T], bool]) -> Callable[[Wa, T], bool]:
    """
    Filter for updates that pass all the given filters.

    >>> all_(text.startswith("Hello"), text.endswith("Word"))
    """
    return lambda wa, m: all(f(wa, m) for f in filters)


def any_(*filters: Callable[[Wa, T], bool]) -> Callable[[Wa, T], bool]:
    """
    Filter for updates that pass any of the given filters.

    >>> any_(text.contains("Hello"), text.regex(r"^World"))
    """
    return lambda wa, m: any(f(wa, m) for f in filters)


def not_(fil: Callable[[Wa, T], bool]) -> Callable[[Wa, T], bool]:
    """
    Filter for updates that don't pass the given filter.

    >>> not_(text.contains("Hello"))
    """
    return lambda wa, m: not fil(wa, m)


def from_users(*numbers: str) -> MessageFilterT | CallbackFilterT | MessageStatusFilterT:
    """
    Filter for messages that are sent from the given numbers.

    >>> from_users("+1 555-555-5555", "972123456789")
    """
    only_nums_pattern = re.compile(r"\D")
    numbers = tuple(re.sub(only_nums_pattern, "", n) for n in numbers)
    return lambda _, m: m.from_user.wa_id in numbers


def from_countries(*prefixes: str | int) -> MessageFilterT | CallbackFilterT | MessageStatusFilterT:
    """
    Filter for messages that are sent from the given country codes.

    - See https://countrycode.org/ for a list of country codes.

    It is always recommended to restrict the countries that can use your bot. remember that you pay for
    every conversation that you reply to.

    >>> from_countries("972", "1") # Israel and USA
    """
    codes = tuple(str(p) for p in prefixes)
    return lambda _, m: m.from_user.wa_id.startswith(codes)


class _BaseUpdateFilter(ABC):
    """
    Base class for all filters.
    """

    def __new__(cls, wa: Wa, m: T) -> bool:
        """When instantiated, call the ``any`` method."""
        return cls.any(wa, m)

    @property
    @abstractmethod
    def __message_types__(self) -> tuple[Mt, ...]:
        """The message types that the filter is for."""
        ...

    @staticmethod
    @abstractmethod
    def any(wa: Wa, m: T) -> bool:
        """Filter for all updates of this type."""
        ...

    @classmethod
    def _match_type(cls, m: Msg) -> bool:
        return m.type in cls.__message_types__


class MediaFilter(_BaseUpdateFilter):
    """
    Useful filters for media messages. Alias: ``filters.media``.
    """
    __message_types__ = (Mt.IMAGE, Mt.VIDEO, Mt.AUDIO, Mt.DOCUMENT, Mt.STICKER)

    any: MessageFilterT = lambda _, m: m.has_media
    """
    Filter for all media messages.
        - Same as ``filters.media``.
    
    >>> filters.media.any
    """

    @classmethod
    def mimetypes(cls, *mimetypes: str) -> MessageFilterT:
        """
        Filter for media messages that match any of the given mime types.

        - `\`Supported Media Types\` on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#supported-media-types>`_.

        >>> media.mimetypes("application/pdf", "image/png")
        >>> video.mimetypes("video/mp4")
        >>> audio.mimetypes("audio/mpeg")
        """
        return lambda _, m: cls._match_type(m) and any(
            t == getattr(m, cls.__message_types__[0].value).mime_type for t in mimetypes
        )


media: MessageFilterT | Type[MediaFilter] = MediaFilter


class _MediaWithCaptionFilter(MediaFilter, ABC):
    @classmethod
    def _has_caption(cls, _: Wa, m: Msg) -> bool:
        return cls._match_type(m) and m.caption is not None

    has_caption: MessageFilterT = _has_caption
    """
    Filter for media messages that have a caption.
    
    >>> filters.media.has_caption
    """


class TextFilter(_BaseUpdateFilter):
    """Useful filters for text messages. Alias: ``filters.text``."""

    __message_types__ = (Mt.TEXT,)

    any: MessageFilterT = lambda _, m: m.type == Mt.TEXT
    """
    Filter for all text messages.
        - Same as ``filters.text``.
    
    >>> filters.text.any
    """

    is_command: MessageFilterT = lambda _, m: m.type == Mt.TEXT and m.text.startswith(("!", "/", "#"))
    """
    Filter for text messages that are commands (start with ``!``, ``/``, or ``#``).
        - Use TextFilter.command if you want to filter for specific commands or prefixes.
    
    >>> filters.text.is_command
    """

    @staticmethod
    def matches(*matches: str, ignore_case: bool = False) -> MessageFilterT:
        """
        Filter for text messages that match exactly the given text/s.

        >>> text.matches("Hello","Hi")

        Args:
            *matches: The text/s to filter for.
            ignore_case: Whether to ignore case when matching (default: ``False``).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, m: TextFilter._match_type(m) and (m.text.lower() if ignore_case else m.text) in matches

    @staticmethod
    def contains(*matches: str, ignore_case: bool = False) -> MessageFilterT:
        """
        Filter for text messages that contain the given text/s.

        >>> text.contains("Cat","Dog",ignore_case=True)

        Args:
            *matches: The text/s to filter for.
            ignore_case: Whether to ignore case when matching. (default: ``False``).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, m: TextFilter._match_type(m) and any(
            t in (m.text.lower() if ignore_case else m.text) for t in matches
        )

    @staticmethod
    def startswith(*matches: str, ignore_case: bool = False) -> MessageFilterT:
        """
        Filter for text messages that start with the given text/s.

        >>> text.startswith("What", "When", ignore_case=True)

        Args:
            *matches: The text/s to filter for.
            ignore_case: Whether to ignore case when matching (default: ``False``).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, m: TextFilter._match_type(m) and (m.text.lower() if ignore_case else m.text).startswith(
            matches)

    @staticmethod
    def endswith(*matches: str, ignore_case: bool = False) -> MessageFilterT:
        """
        Filter for text messages that end with the given text/s.

        >>> text.endswith("Bye", "See you", ignore_case=True)

        Args:
            *matches: The text/s to filter for.
            ignore_case: Whether to ignore case when matching (default: ``False``).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, m: TextFilter._match_type(m) and (m.text.lower() if ignore_case else m.text).endswith(matches)

    @staticmethod
    def regex(*patterns: str | re.Pattern, flags: int = 0) -> MessageFilterT:
        """
        Filter for text messages that match the given regex/regexes.

        >>> text.regex(r"Hello\s+World", r"Bye\s+World", flags=re.IGNORECASE)

        Args:
            *patterns: The regex/regexes to filter for.
            flags: The regex flags to use (default: ``0``).
        """
        patterns = tuple(re.compile(p, flags) if isinstance(p, str) else p for p in patterns)
        return lambda _, m: TextFilter._match_type(m) and any(re.match(p, m.text, flags) for p in patterns)

    @staticmethod
    def length(*lengths: tuple[int, int]) -> MessageFilterT:
        """
        Filter for text messages that have a length between the given range/s.

        >>> text.length((1, 10), (50, 100))

        Args:
            *lengths: The length range/s to filter for (e.g. (1, 10), (50, 100)).
        """
        return lambda _, m: TextFilter._match_type(m) and any(i[0] <= len(m.text) <= i[1] for i in lengths)

    @staticmethod
    def command(*cmds: str, prefixes: str | Iterable[str] = "!", ignore_case: bool = False) -> MessageFilterT:
        """
        Filter for text messages that are commands.

        >>> text.command("start", "hello", prefixes=("!", "/"), ignore_case=True)

        Args:
            *cmds: The command/s to filter for (e.g. "start", "hello").
            prefixes: The prefix/s to filter for (default: "!", i.e. "!start").
            ignore_case: Whether to ignore case when matching (default: ``False``).
        """
        cmds = tuple(c.lower() for c in cmds) if ignore_case else cmds
        return lambda _, m: TextFilter._match_type(m) and (
                m.text[0] in prefixes and (m.text[1:].lower() if ignore_case else m.text[1:]).startswith(cmds)
        )


text: MessageFilterT | Type[TextFilter] = TextFilter


class ImageFilter(_MediaWithCaptionFilter):
    """Useful filters for image messages. Alias: ``filters.image``."""

    __message_types__ = (Mt.IMAGE,)

    any: MessageFilterT = lambda _, m: ImageFilter._match_type(m)
    """
    Filter for all image messages.
        - Same as ``filters.image``.
    
    >>> filters.image.any
    """


image: MessageFilterT | Type[ImageFilter] = ImageFilter


class VideoFilter(_MediaWithCaptionFilter):
    """Useful filters for video messages. Alias: ``filters.video``."""

    __message_types__ = (Mt.VIDEO,)

    any: MessageFilterT = lambda _, m: VideoFilter._match_type(m)
    """
    Filter for all video messages.
        - Same as ``filters.video``.
    
    >>> filters.video.any
    """


video: MessageFilterT | Type[VideoFilter] = VideoFilter


class DocumentFilter(_MediaWithCaptionFilter):
    """Useful filters for document messages. Alias: ``filters.document``."""

    __message_types__ = (Mt.DOCUMENT,)

    any: MessageFilterT = lambda _, m: DocumentFilter._match_type(m)
    """
    Filter for all document messages.
        - Same as ``filters.document``.
    
    >>> filters.document.any
    """


document: MessageFilterT | Type[DocumentFilter] = DocumentFilter


class AudioFilter(MediaFilter):
    """Useful filters for audio messages. Alias: ``filters.audio``."""

    __message_types__ = (Mt.AUDIO,)

    any: MessageFilterT = lambda _, m: AudioFilter._match_type(m)
    """
    Filter for all audio messages (voice notes and audio files).
        - Same as ``filters.audio``.
    
    >>> filters.audio.any
    """

    voice: MessageFilterT = lambda _, m: AudioFilter._match_type(m) and m.audio.voice
    """
    Filter for audio messages that are voice notes.
    
    >>> filters.audio.voice
    """

    audio: MessageFilterT = lambda _, m: AudioFilter._match_type(m) and not m.audio.voice
    """
    Filter for audio messages that are audio files.
    
    >>> filters.audio.audio
    """


audio: MessageFilterT | Type[AudioFilter] = AudioFilter


class StickerFilter(MediaFilter):
    """Useful filters for sticker messages. Alias: ``filters.sticker``."""

    __message_types__ = (Mt.STICKER,)

    any: MessageFilterT = lambda _, m: StickerFilter._match_type(m)
    """
    Filter for all sticker messages.
        - Same as ``filters.sticker``.
    
    >>> filters.sticker.any
    """

    animated: MessageFilterT = lambda _, m: StickerFilter._match_type(m) and m.sticker.animated
    """
    Filter for animated sticker messages.
    
    >>> filters.sticker.animated
    """

    static: MessageFilterT = lambda _, m: StickerFilter._match_type(m) and not m.sticker.animated
    """
    Filter for static sticker messages.
    
    >>> filters.sticker.static
    """


sticker: MessageFilterT | Type[StickerFilter] = StickerFilter


class LocationFilter(_BaseUpdateFilter):
    """Useful filters for location messages. Alias: ``filters.location``."""

    __message_types__ = (Mt.LOCATION,)

    any: MessageFilterT = lambda _, m: LocationFilter._match_type(m)
    """
    Filter for all location messages.
        - Same as ``filters.location``.
    
    >>> filters.location.any
    """

    current_location: MessageFilterT = lambda _, m: LocationFilter._match_type(m) and m.location.current_location
    """
    Filter for location messages that are the current location of the user and not just selected manually.
    
    >>> filters.location.current_location
    """

    @staticmethod
    def in_radius(lat: float, lon: float, radius: float | int) -> MessageFilterT:
        """
        Filter for location messages that are in a given radius.

        >>> location.in_radius(lat=37.48508108998884, lon=-122.14744733542707, radius=1)

        Args:
            lat: Latitude of the center of the radius.
            lon: Longitude of the center of the radius.
            radius: Radius in kilometers.
        """

        def _in_radius(_: Wa, msg: Msg) -> bool:
            return LocationFilter._match_type(msg) and \
                utils.get_distance(
                    lat1=lat, lon1=lon, lat2=msg.location.latitude, lon2=msg.location.longitude
                ) <= radius

        return _in_radius


location: MessageFilterT | Type[LocationFilter] = LocationFilter


class ReactionFilter(_BaseUpdateFilter):
    """Useful filters for reaction messages. Alias: ``filters.reaction``."""

    __message_types__ = (Mt.REACTION,)

    any: MessageFilterT = lambda _, m: ReactionFilter._match_type(m)
    """
    Filter for all reaction updates (added or removed).
        - Same as ``filters.reaction``.
    
    >>> filters.reaction.any
    """

    added: MessageFilterT = lambda _, m: ReactionFilter._match_type(m) and m.reaction.emojis is not None
    """
    Filter for reaction messages that were added.
    
    >>> filters.reaction.added
    """

    removed: MessageFilterT = lambda _, m: ReactionFilter._match_type(m) and m.reaction.emojis is None
    """
    Filter for reaction messages that were removed.
    
    >>> filters.reaction.removed
    """

    @staticmethod
    def emojis(*emojis: str) -> MessageFilterT:
        """
        Filter for custom reaction messages. pass emojis as strings.

        >>> reaction.emojis("👍","👎")
        """
        return lambda _, m: ReactionFilter._match_type(m) and m.reaction.emoji in emojis


reaction: MessageFilterT | Type[ReactionFilter] = ReactionFilter


class ContactsFilter(_BaseUpdateFilter):
    """Useful filters for contact messages. Alias: ``filters.contacts``."""

    __message_types__ = (Mt.CONTACTS,)

    any: MessageFilterT = lambda _, m: ContactsFilter._match_type(m)
    """
    Filter for all contacts messages.
        - Same as ``filters.contacts``.
    
    >>> filters.contacts.any
    """

    has_wa: MessageFilterT = lambda _, m: ContactsFilter._match_type(m) and (
        any((p.wa_id for p in (phone for contact in m.contacts for phone in contact.phones)))
    )
    """
    Filter for contacts messages that have a WhatsApp account.
    
    >>> filters.contacts.has_wa
    """

    @staticmethod
    def count(min_count: int, max_count: int) -> MessageFilterT:
        """
        Filter for contacts messages that have a number of contacts between min_count and max_count.

        >>> contacts.count(1, 1) # ensure only 1 contact
        >>> contacts.count(1, 5) # between 1 and 5 contacts
        """
        return lambda _, m: ContactsFilter._match_type(m) and min_count <= len(m.contacts) <= max_count

    @staticmethod
    def phones(*phones: str) -> MessageFilterT:
        """
        Filter for contacts messages that have the given phone number/s.

        >>> contacts.phones("+1 555-555-5555","972123456789")
        """
        only_nums_pattern = re.compile(r"\D")
        phones = [re.sub(only_nums_pattern, "", p) for p in phones]
        return lambda _, m: ContactsFilter._match_type(m) and (
            any(re.sub(only_nums_pattern, "", p.phones) in phones for contact in m.contacts for p in contact.phones)
        )


contacts: MessageFilterT | Type[ContactsFilter] = ContactsFilter


class OrderFilter(_BaseUpdateFilter):
    """Useful filters for order messages. Alias: ``filters.order``."""

    __message_types__ = (Mt.ORDER,)

    any: MessageFilterT = lambda _, m: OrderFilter._match_type(m)
    """
    Filter for all order messages.
        - Same as ``filters.order``.
    
    >>> filters.order.any
    """

    @staticmethod
    def price(min_price: float, max_price: float) -> MessageFilterT:
        """
        Filter for order messages that have a total price between min_price and max_price.

        Args:
            min_price: Minimum price.
            max_price: Maximum price.

        >>> order.price(1, 100) # total price between 1 and 100
        """
        return lambda _, m: OrderFilter._match_type(m) and min_price <= m.order.total_price <= max_price

    @staticmethod
    def count(min_count: int, max_count: int) -> MessageFilterT:
        """
        Filter for order messages that have a number of items between min_count and max_count.

        Args:
            min_count: Minimum number of items.
            max_count: Maximum number of items.

        >>> order.count(1, 5) # between 1 and 5 items
        """
        return lambda _, m: OrderFilter._match_type(m) and min_count <= len(m.order.products) <= max_count

    @staticmethod
    def has_product(*skus: str) -> MessageFilterT:
        """
        Filter for order messages that have the given product/s.

        Args:
            *skus: The products SKUs.

        >>> order.has_product("pizza_1","pizza_2")
        """
        return lambda _, m: OrderFilter._match_type(m) and (
            any(p.sku in skus for p in m.order.products)
        )


order: MessageFilterT | Type[OrderFilter] = OrderFilter


class UnsupportedMsgFilter(_BaseUpdateFilter):
    """Useful filters for unsupported messages. Alias: ``filters.unsupported``."""

    __message_types__ = (Mt.UNSUPPORTED,)

    def __new__(cls):
        return cls.any

    any: MessageFilterT = lambda _, m: m.type == Mt.UNSUPPORTED
    """
    Filter for all unsupported messages.
        - Same as ``filters.unsupported``.
    
    >>> filters.unsupported.any
    """


unsupported: MessageFilterT | Type[UnsupportedMsgFilter] = UnsupportedMsgFilter


class CallbackFilter(_BaseUpdateFilter):
    """Useful filters for callback queries. Alias: ``filters.callback``."""

    __message_types__ = (Mt.INTERACTIVE,)

    def __new__(cls):
        return cls.any

    any: CallbackFilterT = lambda _, __: True
    """
    Filter for all callback queries (the default).
        - Same as ``filters.callback``.
    
    >>> filters.callback.any
    """

    @staticmethod
    def data_matches(*matches: str, ignore_case: bool = False) -> CallbackFilterT:
        """
        Filter for callbacks their data match exactly the given string/s.

        >>> callback.data_matches("menu")
        >>> callback.data_matches("back","return","exit")

        Args:
            *matches: The string/s to match.
            ignore_case: Whether to ignore case when matching (default: False).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, c: (c.data.lower() if ignore_case else c.data) in matches

    @staticmethod
    def data_startswith(*matches: str, ignore_case: bool = False) -> CallbackFilterT:
        """
        Filter for callbacks their data starts with the given string/s.

        >>> callback.data_startswith("id:")

        Args:
            *matches: The string/s to match.
            ignore_case: Whether to ignore case when matching (default: False).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, c: (c.data.lower() if ignore_case else c.data).startswith(matches)

    @staticmethod
    def data_endswith(*matches: str, ignore_case: bool = False) -> CallbackFilterT:
        """
        Filter for callbacks their data ends with the given string/s.

        >>> callback.data_endswith(":true", ":false")

        Args:
            *matches: The string/s to match.
            ignore_case: Whether to ignore case when matching (default: False).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, c: (c.data.lower() if ignore_case else c.data).endswith(matches)

    @staticmethod
    def data_contains(*matches: str, ignore_case: bool = False) -> CallbackFilterT:
        """
        Filter for callbacks their data contains the given string/s.

        >>> callback.data_contains("back")

        Args:
            *matches: The string/s to match.
            ignore_case: Whether to ignore case when matching (default: False).
        """
        matches = tuple(m.lower() for m in matches) if ignore_case else matches
        return lambda _, c: any((m in (c.data.lower() if ignore_case else c.data) for m in matches))

    @staticmethod
    def data_regex(*patterns: str | re.Pattern, flags: int = 0) -> CallbackFilterT:
        """
        Filter for callbacks their data matches the given regex/regexes.

        >>> callback.data_regex(r"^\d+$")  # only digits

        Args:
            *patterns: The regex/regexes to match.
            flags: The regex flags to use (default: 0).
        """
        patterns = tuple(re.compile(p) if isinstance(p, str) else p for p in patterns)
        return lambda _, c: any((re.match(p, c.data, flags) for p in patterns))


callback: CallbackFilterT | Type[CallbackFilter] = CallbackFilter


class MessageStatusFilter(_BaseUpdateFilter):
    """Useful filters for message status updates. Alias: ``filters.message_status``."""

    __message_types__ = ()

    any: MessageStatusFilterT = lambda _, __: True
    """
    Filter for all message status updates (the default).
        - Same as ``filters.message_status``.
        
    >>> filters.message_status.any
    """

    sent: MessageStatusFilterT = lambda _, s: s.status == Mst.SENT
    """
    Filter for messages that have been sent.
    
    >>> filters.message_status.sent
    """

    delivered: MessageStatusFilterT = lambda _, s: s.status == Mst.DELIVERED
    """
    Filter for messages that have been delivered.
    
    >>> filters.message_status.delivered
    """

    read: MessageStatusFilterT = lambda _, s: s.status == Mst.READ
    """
    Filter for messages that have been read.
    
    >>> filters.message_status.read
    """

    failed: MessageStatusFilterT = lambda _, s: s.status == Mst.FAILED
    """
    Filter for status updates of messages that have failed to send.
    
    >>> filters.message_status.failed
    """

    @staticmethod
    def failed_with(*errors: Type[WhatsAppError] | int) -> MessageStatusFilterT:
        """
        Filter for status updates of messages that have failed to send with the given error/s.

        Args:
            *errors: The exceptions from :mod:`pywa.errors` or error codes to match.

        >>> message_status.failed_with(ReEngagementMessage)
        >>> message_status.failed_with(131051)
        """
        error_codes = tuple(c for c in errors if isinstance(c, int))
        exceptions = tuple(e for e in errors if e not in error_codes and issubclass(e, WhatsAppError))
        return lambda _, s: s.status == Mst.FAILED and (
            any((isinstance(s.error, e) for e in exceptions)) or s.error.error_code in error_codes
        )


message_status: MessageStatusFilterT | Type[MessageStatusFilter] = MessageStatusFilter


class TemplateStatusFilter(_BaseUpdateFilter):
    """Useful filters for template status updates. Alias: ``filters.template_status``."""

    __message_types__ = ()

    any: TemplateStatusFilterT = lambda _, __: True
    """
    Filter for all template status updates (the default).
        - Same as ``filters.template_status``.
        
    >>> filters.template_status.any
    """

    template_name: lambda name: TemplateStatusFilterT = lambda name: lambda _, s: s.template_name == name
    """
    Filter for template status updates that are for the given template name.
    
    >>> template_status.template_name("my_template")
    """

    @staticmethod
    def on_event(*events: Ts.TemplateEvent) -> TemplateStatusFilterT:
        """
        Filter for template status updates that are for the given event/s.

        Args:
            *events: The template events to filter for.

        >>> template_status.on_event(Ts.TemplateEvent.APPROVED)
        """
        return lambda _, s: s.event in events

    @staticmethod
    def on_rejection_reason(*reasons: Ts.TemplateRejectionReason) -> TemplateStatusFilterT:
        """
        Filter for template status updates that are for the given reason/s.

        Args:
            *reasons: The template reasons to filter for.

        >>> template_status.on_rejection_reason(Ts.TemplateRejectionReason.INCORRECT_CATEGORY)
        """
        return lambda _, s: s.reason in reasons


template_status: TemplateStatusFilterT | Type[TemplateStatusFilter] = TemplateStatusFilter



__all__ = [
    'CallbackButton',
    'CallbackSelection',
    'Button',
    'SectionRow',
    'Section',
    'SectionList',
    'CallbackData',
    'CallbackDataT'
]

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Iterable, TypeVar, Generic, Callable, Any
from .base_update import BaseUserUpdate
from .others import Metadata, User, ReplyToMessage, MessageType

if TYPE_CHECKING:
    from pywa.client import WhatsApp


class CallbackData:
    """
    Base class for all callback data classes. Subclass this class to create a type-safe callback data class.

        If you use ``dataclasses``, which is the recommended way (You get free ordered ``__init__`` and extra features),
        you should not use ``kw_only=True``.
        This is because we are limited to 200 characters in the callback data, so we need to use positional arguments.
        So object like ``User(id=123, name='John')`` will be converted to ``123:John``.

        Currently, the following types are supported:
        :class:`str`, :class:`int`, :class:`bool`, :class:`float` and :class:`Enum` that
        inherits from :class:`str` (e.g ``class State(str, Enum)``). You probably won't need more than that to pass
        state and context in the program.

        The characters ``:`` and ``;`` cannot be used when sending callbacks, because they are used as separators.
        You can change the separators by overriding ``__callback_data_sep__`` (for individual objects) and
        ``CallbackData.__callback_sep__`` (In the base class level, affects all child classes).

        When providing subclassed ``CallbackData`` as a ``factory`` parameter in callback handlers, a basic matching
        filter (``startswith(callback_id)``) will be added automatically. So no need to create one yourself.


    Example:

        >>> from pywa import WhatsApp
        >>> from pywa.types import CallbackData, Button
        >>> from dataclasses import dataclass # Use dataclass to get free ordered __init__
        >>> @dataclass(frozen=True, slots=True) # Do not use kw_only=True
        >>> class UserData(CallbackData): # Subclass CallbackData
        ...     id: int
        ...     admin: bool

        >>> wa = WhatsApp(...)
        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user',
        ...     buttons=[Button(title='Get user', callback_data=UserData(id=123, admin=True))]
        ... )

        >>> @wa.on_callback_button(factory=UserData) # Use the factory parameter to convert the callback data
        ... def on_user_data(client: WhatsApp, btn: CallbackButton[UserData]): # For autocomplete
        ...    if btn.data.admin: print(btn.data.id) # Access the data object as an attribute
    """
    __callback_id__: int = 0
    """Unique ID for each callback data class. Do not override this."""
    __callback_sep__: str = ';'
    """The separator between multiple callback objects. (e.g ``123:John;456:Jane``)"""
    __callback_data_sep__: str = ':'
    """The separator between the callback fields, Can be overridden in subclasses (e.g ``123:John``)."""
    __allowed_types__: tuple[type, ...] = (str, int, bool, float)
    """The allowed types in the callback data."""

    def __init_subclass__(cls, **kwargs):
        """Validate the callback data class and set a unique ID for it."""
        super().__init_subclass__(**kwargs)
        if not (annotations := cls.__annotations__.items()):
            raise TypeError(f"Callback data class {cls.__name__} must have at least one field.")
        if unsupported_fields := {
            (field_name, field_type) for field_name, field_type in annotations
            if not issubclass(field_type, (types := cls.__allowed_types__))
        }:
            raise TypeError(f"Unsupported types {unsupported_fields} in callback data. Use one of {types}.")
        cls.__callback_id__ = CallbackData.__callback_id__
        CallbackData.__callback_id__ += 1

    @classmethod
    def from_str(
            cls,
            data: str,
    ) -> 'CallbackData':
        """
        Internal function to convert a callback string to a callback object.
        """
        try:
            # noinspection PyArgumentList
            return cls(*(
                annotation(value) for annotation, value in zip(
                    cls.__annotations__.values(),
                    data.split(cls.__callback_data_sep__)[1:],
                    strict=True
                )
            ))
        except (ValueError, TypeError) as e:
            raise ValueError(f"Invalid callback data for {cls.__name__}: {data}") from e

    @staticmethod
    def _not_contains(value: Any, *not_) -> str:
        """Internal function to validate that the value does not contain the separator."""
        if any(sep in (str_val := str(value)) for sep in not_):
            raise ValueError(f"Callback data cannot contain the characters {not_} "
                             f"Because they are used as separators. \nYou can change the separators by overriding "
                             f"``__callback_data_sep__`` (for individual objects) and ``CallbackData.__callback_sep__`` "
                             f"(In the base class level, affects all child classes).")
        return str_val

    def to_str(self) -> str:
        """
        Internal function to convert a callback object to a callback string.
        """
        return self.__callback_data_sep__.join((str(self.__callback_id__), *(
            self._not_contains(getattr(self, field_name), self.__callback_sep__, self.__callback_data_sep__)
            if not issubclass(field_type, (bool, Enum)) else (' ' if getattr(self, field_name) else '')
            if field_type is bool else self._not_contains(
                getattr(self, field_name).value, self.__callback_sep__, self.__callback_data_sep__
            )
            for field_name, field_type in self.__annotations__.items()
        )))

    @classmethod
    def join_to_str(cls, *datas: Any) -> str:
        """Internal function to join multiple callback objects to a callback string."""
        return cls.__callback_sep__.join(
            data.to_str() if isinstance(data, CallbackData)
            else cls._not_contains(data, cls.__callback_data_sep__)
            for data in datas
        )


CallbackDataT = TypeVar('CallbackDataT', bound=CallbackData | Iterable[type[CallbackData]] | str | Callable[[str], Any])
"""Type hint for `factory` parameter in callback handlers."""


@dataclass(frozen=True, slots=True, kw_only=True)
class CallbackButton(BaseUserUpdate, Generic[CallbackDataT]):
    """
    Represents a callback button.

    Attributes:
        id: The ID of the message.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (always ``INTERACTIVE``).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent.
        reply_to_message: The message to which this callback button is a reply to.
        data: The data of the button (the ``callback_data`` parameter in :class:`Button`).
        title: The title of the button.
    """
    id: str
    type: MessageType
    metadata: Metadata
    from_user: User
    timestamp: datetime
    reply_to_message: ReplyToMessage
    data: CallbackDataT
    title: str

    @property
    def message_id_to_reply(self) -> str:
        """The ID of the message to reply to."""
        return self.reply_to_message.message_id

    @classmethod
    def from_update(cls, client: 'WhatsApp', update: dict) -> 'CallbackButton':
        msg = (value := update['entry'][0]['changes'][0]['value'])['messages'][0]
        return cls(
            _client=client,
            id=msg['id'],
            metadata=Metadata.from_dict(value['metadata']),
            type=MessageType(msg['type']),
            from_user=User.from_dict(value['contacts'][0]),
            timestamp=datetime.fromtimestamp(int(msg['timestamp'])),
            reply_to_message=ReplyToMessage.from_dict(msg['context']),
            data=msg['interactive']['button_reply']['id'],
            title=msg['interactive']['button_reply']['title']
        )


@dataclass(frozen=True, slots=True, kw_only=True)
class CallbackSelection(BaseUserUpdate, Generic[CallbackDataT]):
    """
    Represents a callback selection.

    Attributes:
        id: The ID of the message.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (always ``INTERACTIVE``).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent.
        reply_to_message: The message to which this callback selection is a reply to.
        data: The data of the selection (the ``callback_data`` parameter in :class:`SectionRow`).
        title: The title of the selection.
        description: The description of the selection (optional).
    """
    id: str
    type: MessageType
    metadata: Metadata
    from_user: User
    timestamp: datetime
    reply_to_message: ReplyToMessage
    data: CallbackDataT
    title: str
    description: str | None

    @classmethod
    def from_update(cls, client: 'WhatsApp', update: dict) -> 'CallbackSelection':
        msg = (value := update['entry'][0]['changes'][0]['value'])['messages'][0]
        return cls(
            _client=client,
            id=msg['id'],
            metadata=Metadata.from_dict(value['metadata']),
            type=MessageType(msg['type']),
            from_user=User.from_dict(value['contacts'][0]),
            timestamp=datetime.fromtimestamp(int(msg['timestamp'])),
            reply_to_message=ReplyToMessage.from_dict(msg['context']),
            data=msg['interactive']['list_reply']['id'],
            title=msg['interactive']['list_reply']['title'],
            description=msg['interactive']['list_reply'].get('description')
        )


@dataclass(frozen=True, slots=True)
class Button:
    """
    Represents a button in the button list.

    Attributes:
        title: The title of the button (up to 20 characters).
        callback_data: The data to send when the user clicks on the button (up to 256 characters, for complex data
         You can use :class:`CallbackData`).
    """
    title: str
    callback_data: CallbackDataT

    def to_dict(self) -> dict:
        return {
            "type": "reply",
            "reply":
                {
                    "id": self.callback_data.to_str()
                    if isinstance(self.callback_data, CallbackData) else self.callback_data
                    if isinstance(self.callback_data, str) else self.callback_data.join_to_str(*self.callback_data),
                    "title": self.title
                }
        }


@dataclass(frozen=True, slots=True)
class SectionRow:
    """
    Represents a row in a section.

    Attributes:
        title: The title of the row (up to 24 characters).
        callback_data: The payload to send when the user clicks on the row (up to 200 characters, for complex data
            You can use :class:`CallbackData`).
        description: The description of the row (optional, up to 72 characters).
    """
    title: str
    callback_data: CallbackDataT
    description: str | None = None

    def to_dict(self) -> dict:
        d = {
            "id": self.callback_data.to_str() if isinstance(self.callback_data, CallbackData)
            else self.callback_data if isinstance(self.callback_data, str)
            else self.callback_data.join_to_str(*self.callback_data),
            "title": self.title
        }
        if self.description:
            d["description"] = self.description
        return d


@dataclass(frozen=True, slots=True)
class Section:
    """
    Represents a section in a section list.

    Attributes:
        title: The title of the section (up to 24 characters).
        rows: The rows in the section (at least 1, no more than 10).
    """
    title: str
    rows: Iterable[SectionRow]

    def to_dict(self) -> dict:
        return {
            "title": self.title,
            "rows": tuple(row.to_dict() for row in self.rows)
        }


@dataclass(frozen=True, slots=True)
class SectionList:
    """
    Represents a section list in an interactive message.

    Attributes:
        button_title: The title of the button that opens the section list (up to 20 characters).
        sections: The sections in the section list (at least 1, no more than 10).
    """
    button_title: str
    sections: Iterable[Section]

    def to_dict(self) -> dict:
        return {
            "button": self.button_title,
            "sections": tuple(section.to_dict() for section in self.sections)
        }



import pytest
import dataclasses
from pywa.types import CallbackData


def test_callback_class_uniqueness():
    """Test the uniqueness of the callback data."""
    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: str
        name: str
        is_admin: bool

    @dataclasses.dataclass(slots=True, frozen=True)
    class Group(CallbackData):
        id: str
        name: str

    assert User.__callback_id__ != Group.__callback_id__, 'The callback id must be unique for each child class.'


def test_callback_class_not_empty():
    """Test the callback data."""
    with pytest.raises(TypeError):

        @dataclasses.dataclass(slots=True, frozen=True)
        class User(CallbackData):
            pass

    with pytest.raises(TypeError):
        @dataclasses.dataclass(slots=True, frozen=True)
        class User(CallbackData):
            id: str
            name: str
            is_admin: bool
            x: int

        @dataclasses.dataclass(slots=True, frozen=True)
        class UserChild(User):
            pass


def test_callback_supported_types():
    with pytest.raises(TypeError):
        @dataclasses.dataclass(slots=True, frozen=True)
        class User(CallbackData):
            id: int
            name: str
            is_admin: bool
            x: dict
            y: list
            z: set


def test_data():
    """Test the callback data."""
    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: int
        name: str
        is_admin: bool

    assert User(id=1234, name='xxx', is_admin=True).to_str() == '7:1234:xxx: '
    assert User(id=3456, name='yyy', is_admin=False).to_str() == '7:3456:yyy:'
    assert User.from_str('1:1234:xxx: ') == User(id=1234, name='xxx', is_admin=True)
    assert User.from_str('1:3456:yyy:') == User(id=3456, name='yyy', is_admin=False)

    with pytest.raises(ValueError):
        User.from_str('1:3456')

    with pytest.raises(ValueError):
        User.from_str('1:3456:yyy: :')

    with pytest.raises(ValueError):
        User.from_str('1:x:yyy: ')


def test_multiple_data():
    """Test the callbacks separator."""
    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: str
        name: str
        is_admin: bool

    @dataclasses.dataclass(slots=True, frozen=True)
    class Group(CallbackData):
        id: str
        name: str

    user = User(id='1234', name='xxx', is_admin=True)
    group = Group(id='3456', name='yyy')
    assert CallbackData.join_to_str(user, group) == '9:1234:xxx: ;11:3456:yyy'
    assert CallbackData.join_to_str(user, group, user) == '9:1234:xxx: ;11:3456:yyy;9:1234:xxx: '
    x, y = '3:1234:xxx: ;5:3456:yyy'.split(CallbackData.__callback_sep__)
    assert (User.from_str(x), Group.from_str(y)) == (user, group)
    x, y, z = '3:1234:xxx: ;5:3456:yyy;3:1234:xxx: '.split(CallbackData.__callback_sep__)
    assert (User.from_str(x), Group.from_str(y), User.from_str(z)) == (user, group, user)


def test_data_sep():
    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        __callback_data_sep__ = '*'
        id: int
        name: str
        is_admin: bool

    assert User(id=1234, name='xxx', is_admin=True).to_str() == '13*1234*xxx* '
    assert User(id=3456, name='yyy', is_admin=False).to_str() == '13*3456*yyy*'

    with pytest.raises(ValueError):
        User.from_str('1:3456:David*Lev:')

    try:
        User.from_str('1*3456*David:Lev*')
    except ValueError:
        pytest.fail('The data separator override does not work.')


def test_data_sep_multiple_data():
    CallbackData.__callback_sep__ = '^'

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: int
        name: str
        is_admin: bool

    @dataclasses.dataclass(slots=True, frozen=True)
    class Group(CallbackData):
        id: int
        name: str

    user = User(id=1234, name='xxx', is_admin=True)
    group = Group(id=3456, name='yyy')
    assert CallbackData.join_to_str(user, group) == '15:1234:xxx: ^17:3456:yyy'
    assert CallbackData.join_to_str(user, group, user) == '15:1234:xxx: ^17:3456:yyy^15:1234:xxx: '
    x, y = '3:1234:xxx: ^5:3456:yyy'.split(CallbackData.__callback_sep__)
    assert (User.from_str(x), Group.from_str(y)) == (user, group)
    x, y, z = '3:1234:xxx: ^5:3456:yyy^3:1234:xxx: '.split(CallbackData.__callback_sep__)
    assert (User.from_str(x), Group.from_str(y), User.from_str(z)) == (user, group, user)






from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Callable, Any, Iterable
from pywa.errors import WhatsAppError
from .callback import Button, SectionList
from .base_update import BaseUserUpdate
from .others import ReplyToMessage, Reaction, Location, Contact, User, Metadata, Order, System, MessageType, \
    ProductsSection
from .media import Image, Video, Sticker, Document, Audio

if TYPE_CHECKING:
    from pywa.client import WhatsApp


_FIELDS_TO_OBJECTS_CONSTRUCTORS: dict[str, Callable[[dict, WhatsApp], Any]] = dict(
    text=lambda m, _client: m['body'],
    image=Image.from_dict,
    video=Video.from_dict,
    sticker=Sticker.from_dict,
    document=Document.from_dict,
    audio=Audio.from_dict,
    reaction=Reaction.from_dict,
    location=Location.from_dict,
    contacts=lambda m, _client: tuple(Contact.from_dict(c) for c in m),
    order=Order.from_dict,
    system=System.from_dict,
)

_MEDIA_FIELDS = ('image', 'video', 'sticker', 'document', 'audio')


@dataclass(frozen=True, slots=True, kw_only=True)
class Message(BaseUserUpdate):
    """
    A message received from a user.

    - `\`Message\` on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/components#messages-object>`_

    Attributes:
        id: The message ID (If you want to reply to the message, use ``message_id_to_reply`` instead).
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (See :class:`MessageType`).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent.
        reply_to_message: The message to which this message is a reply to. (Optional)
        forwarded: Whether the message was forwarded.
        forwarded_many_times: Whether the message was forwarded many times.
         (when True, ``forwarded`` will be True as well)
        text: The text of the message (if the message type is :class:`MessageType.TEXT`).
        image: The image of the message (if the message type is :class:`MessageType.IMAGE`).
        video: The video of the message (if the message type is :class:`MessageType.VIDEO`).
        sticker: The sticker of the message (if the message type is :class:`MessageType.STICKER`).
        document: The document of the message (if the message type is :class:`MessageType.DOCUMENT`).
        audio: The audio of the message (if the message type is :class:`MessageType.AUDIO`).
        caption: The caption of the message (Optional, only available for :class:`MessageType.IMAGE`,
         :class:`MessageType.VIDEO` and :class:`MessageType.DOCUMENT`).
        reaction: The reaction of the message (if the message type is :class:`MessageType.REACTION`).
        location: The location of the message (if the message type is :class:`MessageType.LOCATION`).
        contacts: The contacts of the message (if the message type is :class:`MessageType.CONTACTS`).
        order: The order of the message (if the message type is :class:`MessageType.ORDER`).
        system: The system update (if the message type is :class:`MessageType.SYSTEM`).
        error: The error of the message (if the message type is :class:`MessageType.UNSUPPORTED`).
    """
    id: str
    type: MessageType
    metadata: Metadata
    from_user: User
    timestamp: datetime
    reply_to_message: ReplyToMessage | None
    forwarded: bool
    forwarded_many_times: bool
    text: str | None = None
    image: Image | None = None
    video: Video | None = None
    sticker: Sticker | None = None
    document: Document | None = None
    audio: Audio | None = None
    caption: str | None = None
    reaction: Reaction | None = None
    location: Location | None = None
    contacts: tuple[Contact] | None = None
    order: Order | None = None
    system: System | None = None
    error: WhatsAppError | None = None

    @property
    def message_id_to_reply(self) -> str:
        """The ID of the message to reply to."""
        return self.id if self.type != MessageType.REACTION else self.reaction.message_id

    @property
    def has_media(self) -> bool:
        """
        Whether the message has any media. (image, video, sticker, document or audio)
            - If you want to get the media of the message, use :attr:`~Message.media` instead.
        """
        return self.media is not None

    @property
    def is_reply(self) -> bool:
        """
        Whether the message is a reply to another message.

        - Reaction messages are also considered as replies (But ``.reply_to_message`` will be ``None``).
        """
        return self.reply_to_message is not None or self.reaction is not None

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> Message:
        msg = (value := update['entry'][0]['changes'][0]['value'])['messages'][0]
        msg_type = msg['type']
        context = msg.get('context')
        constructor = _FIELDS_TO_OBJECTS_CONSTRUCTORS.get(msg_type)
        msg_content = {msg_type: constructor(msg[msg_type], _client=client)} if constructor is not None else {} # noqa
        return cls(
            _client=client,
            id=msg['id'],
            type=MessageType(msg['type']),
            **msg_content,
            from_user=User.from_dict(value['contacts'][0]),
            timestamp=datetime.fromtimestamp(int(msg['timestamp'])),
            metadata=Metadata.from_dict(value['metadata']),
            forwarded=any(context.get(key) for key in ('forwarded', 'frequently_forwarded')) if context else False,
            forwarded_many_times=context.get('frequently_forwarded', False) if context else False,
            reply_to_message=ReplyToMessage.from_dict(context),
            caption=msg[msg_type].get('caption') if msg_type in ('image', 'video', 'document') else None,
            error=WhatsAppError.from_incoming_error(msg['errors'][0]) if 'errors' in msg else None
        )

    @property
    def media(self) -> Image | Video | Sticker | Document | Audio | None:
        """
        The media of the message if any, otherwise ``None``. (image, video, sticker, document or audio)
            - If you want to check whether the message has any media, use :attr:`~Message.has_media` instead.
        """
        return next((getattr(self, media_type) for media_type in _MEDIA_FIELDS if getattr(self, media_type)), None)

    def download_media(
            self,
            filepath: str | None = None,
            filename: str | None = None,
            in_memory: bool = False,
    ) -> str | bytes:
        """
        Download a media file from WhatsApp servers (image, video, sticker, document or audio).

        Args:
            filepath: The path where to save the file (if not provided, the current working directory will be used).
            filename: The name of the file (if not provided, it will be guessed from the URL + extension).
            in_memory: Whether to return the file as bytes instead of saving it to disk (default: False).

        Returns:
            The path of the saved file if ``in_memory`` is False, the file as bytes otherwise.

        Raises:
            ValueError: If the message does not contain any media.
        """
        try:
            return self.media.download(path=filepath, filename=filename, in_memory=in_memory)
        except AttributeError:
            raise ValueError('Message does not contain any media.')

    def copy(
            self,
            to: str,
            header: str | None = None,
            body: str | None = None,
            footer: str | None = None,
            keyboard: Iterable[Button] | SectionList | None = None,
            preview_url: bool = False,
            reply_to_message_id: str = None,
    ) -> str:
        """
        Copy incoming message to another chat
            - The WhatsApp Cloud API does not offer a `real` forward option so this is just copy the message content.

        Args:
            to: The phone ID of the WhatsApp user to copy the message to.
            header: The header of the message (if keyboard is provided, optional, up to 60 characters, no markdown allowed).
            body: The body of the message (if keyboard are provided, optional, up to 1024 characters, markdown allowed).
            footer: The footer of the message (if keyboard is provided, optional, markdown has no effect).
            keyboard: The buttons to send with the message (only in case of message from type ``text``, ``document``,
             ``video`` and ``image``. also, the ``SectionList`` is only available to ``text`` type)
            reply_to_message_id:  The message ID to reply to (optional).
            preview_url: Whether to show a preview of the URL in the message (if any).

        Returns:
            The ID of the sent message.

        Raises:
            ValueError: If the message type is ``reaction`` and no ``reply_to_message_id`` is provided, or if the message
             type is ``unsupported``.
        """
        match self.type:
            case MessageType.TEXT:
                return self._client.send_message(
                    to=to,
                    text=self.text,
                    preview_url=preview_url,
                    header=header,
                    footer=footer,
                    keyboard=keyboard,
                    reply_to_message_id=reply_to_message_id,
                )
            case MessageType.DOCUMENT:
                return self._client.send_document(
                    to=to,
                    document=self.document.id,
                    filename=self.document.filename,
                    caption=self.caption,
                    body=body,
                    footer=footer,
                    buttons=keyboard,
                    reply_to_message_id=reply_to_message_id,
                )
            case MessageType.IMAGE:
                return self._client.send_image(
                    to=to,
                    image=self.image.id,
                    caption=self.caption,
                    body=body,
                    footer=footer,
                    buttons=keyboard,
                    reply_to_message_id=reply_to_message_id,
                )
            case MessageType.VIDEO:
                return self._client.send_video(
                    to=to,
                    video=self.video.id,
                    caption=self.caption,
                    buttons=keyboard,
                    body=body,
                    footer=footer,
                    reply_to_message_id=reply_to_message_id,
                )
            case MessageType.STICKER:
                return self._client.send_sticker(
                    to=to,
                    sticker=self.sticker.id
                )
            case MessageType.LOCATION:
                return self._client.send_location(
                    to=to,
                    latitude=self.location.latitude,
                    longitude=self.location.longitude,
                    name=self.location.name,
                    address=self.location.address
                )
            case MessageType.AUDIO:
                return self._client.send_audio(
                    to=to,
                    audio=self.audio.id
                )
            case MessageType.CONTACTS:
                return self._client.send_contact(
                    to=to,
                    contact=self.contacts,
                    reply_to_message_id=reply_to_message_id,
                )
            case MessageType.REACTION:
                if reply_to_message_id is None:
                    raise ValueError("You need to provide `reply_to_message_id` in order to `copy` a reaction")
                return self._client.send_reaction(
                    to=to,
                    message_id=reply_to_message_id,
                    emoji=self.reaction.emoji or ""
                )
            case MessageType.ORDER:
                if len(self.order.products) == 1:
                    return self._client.send_product(
                        to=to,
                        catalog_id=self.order.catalog_id,
                        sku=self.order.products[0].sku,
                        body=body,
                        footer=footer,
                        reply_to_message_id=reply_to_message_id,
                    )
                return self._client.send_products(
                    to=to,
                    catalog_id=self.order.catalog_id,
                    product_sections=(ProductsSection(
                        title=header,
                        skus=(p.sku for p in self.order.products)
                    ),),
                    title=header,
                    body=body,
                    footer=footer,
                    reply_to_message_id=reply_to_message_id,
                    )
            case MessageType.SYSTEM:
                return self._client.send_message(
                    to=to,
                    text=self.system.body,
                    header=header,
                    footer=footer,
                    keyboard=keyboard,
                    reply_to_message_id=reply_to_message_id,
                )
            case _:
                raise ValueError(f"Message of type {self.type} cannot be copied.")



"""The internal API for the WhatsApp client."""

import requests
from typing import BinaryIO, Any
from requests_toolbelt import MultipartEncoder
from pywa.errors import WhatsAppError


class WhatsAppCloudApi:
    """Internal methods for the WhatsApp client."""

    def __init__(
            self,
            phone_id: str,
            token: str,
            session: requests.Session,
            base_url: str,
            api_version: float,
    ):
        self.phone_id = phone_id
        if session.headers.get("Authorization") is not None:
            raise ValueError("You can't use the same requests.Session for multiple WhatsApp instances!")
        self._session = session
        self._base_url = f"{base_url}/v{api_version}"
        self._session.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
        }
        self._common_keys = {
            "messaging_product": "whatsapp",
            "recipient_type": "individual",
        }

    def __str__(self) -> str:
        return f"WhatsAppCloudApi(phone_id={self.phone_id!r})"

    def __repr__(self) -> str:
        return self.__str__()

    def _make_request(
            self,
            method: str,
            endpoint: str,
            **kwargs
    ) -> dict | list:
        """
        Internal method to make a request to the WhatsApp API.

        Args:
            method: The HTTP method to use.
            endpoint: The endpoint to request.
            **kwargs: Additional arguments to pass to the request.

        Returns:
            The response JSON.

        Raises:
            WhatsAppError: If the request failed.
        """
        res = self._session.request(method=method, url=f"{self._base_url}{endpoint}", **kwargs)
        if res.status_code >= 400:
            raise WhatsAppError.from_response(status_code=res.status_code, error=res.json()["error"])
        return res.json()

    def send_text_message(
            self,
            to: str,
            text: str,
            preview_url: bool = False,
            reply_to_message_id: str | None = None,
    ) -> dict[str, dict | list]:
        """
        Send a text message to a WhatsApp user.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        Args:
            to: The WhatsApp ID of the recipient.
            text: The text to send.
            preview_url: Whether to show a preview of the URL in the message.
            reply_to_message_id: The ID of the message to reply to.

        Returns:
            The sent message.
        """
        data = {
            **self._common_keys,
            "to": to,
            "type": "text",
            "text": {
                "body": text,
                "preview_url": preview_url
            },
        }
        if reply_to_message_id:
            data["context"] = {"message_id": reply_to_message_id}

        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages",
            json=data
        )

    def upload_media(
            self,
            media: bytes | BinaryIO,
            mime_type: str,
            filename: str,
    ) -> dict[str, str]:
        """
        Upload a media file to WhatsApp.

        Return example::

            {
              'id':'<MEDIA_ID>'
            }

        Args:
            media: media bytes or open(path, 'rb') object
            mime_type: The type of the media file
            filename: The name of the media file
        Returns:
            A dict with the ID of the uploaded media file.
        """
        headers = self._session.headers.copy()
        form_data = MultipartEncoder(
            {
                "file": (
                    filename,
                    media,
                    mime_type
                ),
                "messaging_product": "whatsapp",
                "type": mime_type
            }
        )
        headers["Content-Type"] = form_data.content_type
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/media",
            headers=headers,
            data=form_data
        )

    def get_media_url(self, media_id: str) -> dict:
        """
        Get the URL of a media file.
            - The url is valid for 5 minutes and can be downloaded only with access token.
            - For more info: https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#retrieve-media-url

        Return example::

            {
                'url': 'https://lookaside.fbsbx.com/whatsapp_business/attachments/?mid=645645&ext=54353&hash=xxx-xx',
                'mime_type': 'image/jpeg',
                'sha256': '73298ec14751fhfonf4wfxxxf52f4fb031db3892ff5',
                'file_size': 61901,
                'id': '137524587935346',
                'messaging_product': 'whatsapp'
            }

        Args:
            media_id: The ID of the media file.

        Returns:
            A dict with the URL and other info about the media file.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{media_id}"
        )

    def get_media_bytes(
            self,
            media_url: str,
    ) -> tuple[bytes, str | None]:
        """
        Get the bytes of a media file from WhatsApp servers.

        Args:
            media_url: The URL of the media file (from ``get_media_url``).

        Returns:
            The media file bytes and the MIME type (if available).
        """
        headers = self._session.headers.copy()
        del headers["Content-Type"]
        res = self._session.get(media_url, headers=headers)
        res.raise_for_status()
        return res.content, res.headers.get("Content-Type")

    def delete_media(self, media_id: str) -> dict[str, bool]:
        """
        Delete a media file from WhatsApp servers.

        Return example::

            {'success': True}

        Args:
            media_id: The ID of the media file.

        Returns:
            True if the media file was deleted successfully, False otherwise.
        """
        return self._make_request(
            method="DELETE",
            endpoint=f"/{media_id}"
        )

    def send_media(
            self,
            to: str,
            media_id_or_url: str,
            media_type: str,
            is_url: bool,
            caption: str | None = None,
            filename: str | None = None,
    ) -> dict[str, dict | list]:
        """
        Send a media file to a WhatsApp user.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        Args:
            to: The WhatsApp ID of the recipient.
            media_id_or_url: The ID or URL of the media file to send.
            is_url: Whether the media_id_or_url is a URL or an ID.
            media_type: The type of the media file (e.g. 'image', 'video', 'document').
            caption: The caption to send with the media file (only for images, videos and documents).
            filename: The filename to send with the media file (only for documents).

        Returns:
            The sent message.
        """
        data = {
            **self._common_keys,
            "to": to,
            "type": media_type,
            media_type: {
                ("link" if is_url else "id"): media_id_or_url,
                **({"caption": caption} if caption else {}),
                **({"filename": filename} if filename else {})
            }
        }
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages",
            json=data
        )

    def send_reaction(
            self,
            to: str,
            emoji: str,
            message_id: str,
    ) -> dict[str, dict | list]:
        """
        Send a reaction to a message.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        Args:
            to: The WhatsApp ID of the recipient.
            emoji: The emoji to react with (empty to remove reaction).
            message_id: The ID of the message to react to.

        Returns:
            The sent message.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages/",
            json={
                **self._common_keys,
                "to": to,
                "type": "reaction",
                "reaction": {
                    "emoji": emoji,
                    "message_id": message_id
                }
            }
        )

    def send_location(
            self,
            to: str,
            latitude: float,
            longitude: float,
            name: str | None = None,
            address: str | None = None,
    ) -> dict[str, dict | list]:
        """
        Send a location to a WhatsApp user.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        Args:
            to: The WhatsApp ID of the recipient.
            latitude: The latitude of the location.
            longitude: The longitude of the location.
            name: The name of the location.
            address: The address of the location.

        Returns:
            The sent message.
        """
        data = {
            **self._common_keys,
            "to": to,
            "type": "location",
            "location": {
                "latitude": latitude,
                "longitude": longitude,
                "name": name,
                "address": address
            }
        }

        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages",
            json=data
        )

    def send_raw_request(
            self,
            method: str,
            endpoint: str,
            **kwargs
    ) -> Any:
        """
        Send a raw JSON message to a WhatsApp Cloud API.

        - The endpoint can contain path parameters (e.g. ``/v1/messages/{phone_id}``). only ``phone_id`` is supported.
        - Every request will automatically include the ``Authorization`` and ``Content-Type`` headers. you can override
            them by passing them in ``kwargs`` (headers={...}).

        Args:
            method: The HTTP method to use (e.g. ``POST``, ``GET``, etc.).
            endpoint: The endpoint to send the message to (e.g. ``/v1/messages``).
            **kwargs: Additional arguments to send with the message (e.g. ``json={...}``, headers={...}, etc.).

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method=method,
            endpoint=endpoint.format(phone_id=self.phone_id),
            **kwargs
        )

    def send_interactive_message(
            self,
            to: str,
            type_: str,
            action: dict[str, Any],
            header: dict | None = None,
            body: str | None = None,
            footer: str | None = None,
            reply_to_message_id: str | None = None,
    ) -> dict[str, dict | list]:
        """
        Send an interactive message to a WhatsApp user.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        Args:
            to: The WhatsApp ID of the recipient.
            type_: The type of the message (e.g. ``list``, ``button``, ``product``, etc.).
            action: The action of the message.
            header: The header of the message.
            body: The body of the message.
            footer: The footer of the message.
            reply_to_message_id: The ID of the message to reply to.

        Returns:
            The sent message.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages",
            json={
                **self._common_keys,
                "to": to,
                "type": "interactive",
                "interactive": {
                    "type": type_,
                    "action": action,
                    **({"header": header} if header else {}),
                    **({"body": {"text": body}} if body else {}),
                    **({"footer": {"text": footer}} if footer else {}),
                },
                **({"context": {"message_id": reply_to_message_id}} if reply_to_message_id else {}),
            }
        )

    def send_contacts(
            self,
            to: str,
            contacts: tuple[dict[str, Any]],
            reply_to_message_id: str | None = None,
    ) -> dict[str, dict | list]:
        """
        Send a list of contacts to a WhatsApp user.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        Args:
            to: The WhatsApp ID of the recipient.
            contacts: The contacts to send.
            reply_to_message_id: The ID of the message to reply to.

        Returns:
            The sent message.
        """
        data = {
            **self._common_keys,
            "to": to,
            "type": "contacts",
            "contacts": tuple(contacts)
        }
        if reply_to_message_id:
            data["context"] = {"message_id": reply_to_message_id}
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages",
            json=data
        )

    def mark_message_as_read(
            self,
            message_id: str
    ) -> dict[str, bool]:
        """
        Mark a message as read.

        Return example::

            {
                'success': True
            }

        Args:
            message_id: The ID of the message to mark as read.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages",
            json={
                "messaging_product": "whatsapp",
                "status": "read",
                "message_id": message_id
            }
        )

    def get_business_profile(self) -> dict[str, list[dict[str, str | list[str]]]]:
        """
        Get the business profile.

        Return example::

            {
              "data": [{
                "about": "ABOUT",
                "address": "ADDRESS",
                "description": "DESCRIPTION",
                "email": "EMAIL",
                "messaging_product": "whatsapp",
                "profile_picture_url": "https://URL",
                "websites": [
                   "https://WEBSITE-1",
                   "https://WEBSITE-2"
                 ],
                "vertical": "INDUSTRY",
              }]
            }
        """
        fields = ('about', 'address', 'description', 'email', 'profile_picture_url', 'websites', 'vertical')
        return self._make_request(
            method="GET",
            endpoint=f"/{self.phone_id}/whatsapp_business_profile?fields={','.join(fields)}"
        )

    def update_business_profile(self, data: dict[str, str | list[str]]) -> dict[str, bool]:
        """
        Update the business profile.

        Args:
            data: The data to update the business profile with.

        Return example::

            {
                "success": True
            }
        """
        data.update(messaging_product="whatsapp")
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/whatsapp_business_profile",
            json=data
        )

    def get_commerce_settings(self) -> dict[str, list[dict]]:
        """
        Get the commerce settings of the business catalog.

        Return example::

            {
              "data": [
                {
                  "is_cart_enabled": True,
                  "is_catalog_visible": True,
                  "id": "727705352028726"
                }
              ]
            }
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{self.phone_id}/whatsapp_commerce_settings"
        )

    def update_commerce_settings(self, data: dict) -> dict[str, bool]:
        """
        Change the commerce settings of the business catalog.

        Args:
            data: The data to update the commerce settings with.

        Return example::

            {
              "success": True
            }
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/whatsapp_commerce_settings",
            params=data
        )

    def create_template(
            self,
            business_account_id: str,
            template: dict[str, str | list[str]]
    ) -> dict[str, str]:
        """
        Create a message template.

        Args:
            business_account_id: The ID of the business account.
            template: The template to create.

        Return example::

            {
                "id": "594425479261596",
                "status": "PENDING",
                "category": "MARKETING"
            }
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{business_account_id}/message_templates",
            json=template
        )

    def send_template(
            self,
            to: str,
            template: dict,
            reply_to_message_id: str | None = None
    ) -> dict[str, dict | list]:
        """
        Send a template to a WhatsApp user.

        Args:
            to: The WhatsApp ID of the recipient.
            template: The template to send.
            reply_to_message_id: The ID of the message to reply to.

        Returns example::

            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        """
        data = {
            **self._common_keys,
            "to": to,
            "type": "template",
            "template": template
        }
        if reply_to_message_id:
            data["context"] = {"message_id": reply_to_message_id}
        return self._make_request(
            method="POST",
            endpoint=f"/{self.phone_id}/messages",
            json=data
        )



from __future__ import annotations

from dataclasses import dataclass, fields
from enum import Enum
from importlib import import_module
from typing import Protocol, Any, Callable


def is_fastapi_app(app):
    """Check if the app is a FastAPI app."""
    try:
        return isinstance(app, import_module("fastapi").FastAPI)
    except ImportError:
        return False


def is_flask_app(app):
    """Check if the app is a Flask app."""
    try:
        return isinstance(app, import_module("flask").Flask)
    except ImportError:
        return False


def get_distance(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    """
    Calculate the distance between two points on Earth using the Haversine formula.
    """
    math = import_module("math")
    lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])
    return (2 * math.asin(
        math.sqrt(
            math.sin((lat2 - lat1) / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin((lon2 - lon1) / 2) ** 2)
    )) * 6371


class StrEnum(str, Enum):
    def __str__(self):
        return self.value

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"


@dataclass(frozen=True, slots=True, kw_only=True)
class FromDict:
    """Allows to ignore extra fields when creating a dataclass from a dict."""

    # noinspection PyArgumentList
    @classmethod
    def from_dict(cls, data: dict, **kwargs):
        data.update(kwargs)
        return cls(**{
            k: v for k, v in data.items()
            if k in (f.name for f in fields(cls))
        })


class FastAPIApp(Protocol):
    def get(self, path: str) -> Callable:
        ...

    def post(self, path: str) -> Callable:
        ...


class FlaskApp(Protocol):
    def route(self, rule: str, **options: Any) -> Callable:
        ...




"""
This package contains all the types used in the library.
"""

from .message import Message
from .others import (Contact, User, Reaction, Location, Metadata, ReplyToMessage, Order, Product, System,
                     ProductsSection, CommerceSettings, BusinessProfile, Industry, MessageType)
from .callback import (CallbackButton, CallbackSelection, Button, SectionRow, Section, SectionList, CallbackData)
from .message_status import MessageStatus, MessageStatusType, Conversation, ConversationCategory
from .media import MediaUrlResponse
from .template import NewTemplate, TemplateResponse, Template, TemplateStatus



from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING
from pywa.errors import WhatsAppError
from pywa import utils
from .base_update import BaseUserUpdate
from .others import Metadata, User

if TYPE_CHECKING:
    from pywa.client import WhatsApp


class MessageStatusType(utils.StrEnum):
    """
    Message status type.

    Attributes:
        SENT: The message was sent.
        DELIVERED: The message was delivered.
        READ: The message was read.
        FAILED: The message failed to send (you can access the error with ``.error`` attribute).
    """
    SENT = 'sent'
    DELIVERED = 'delivered'
    READ = 'read'
    FAILED = 'failed'

    @classmethod
    def _missing_(cls, value: str) -> MessageStatusType:
        return cls.FAILED


class ConversationCategory(utils.StrEnum):
    """
    Conversation category.

    Attributes:
        AUTHENTICATION: The conversation is related to authentication.
        MARKETING: The conversation is related to marketing.
        UTILITY: The conversation is related to utility.
        SERVICE: The conversation is related to service.
        REFERRAL_CONVERSION: The conversation is related to referral conversion.
        UNKNOWN: The conversation category is unknown.
    """
    AUTHENTICATION = 'authentication'
    MARKETING = 'marketing'
    UTILITY = 'utility'
    SERVICE = 'service'
    REFERRAL_CONVERSION = 'referral_conversion'
    UNKNOWN = 'unknown'

    @classmethod
    def _missing_(cls, value: str) -> ConversationCategory:
        return cls.UNKNOWN


@dataclass(frozen=True, slots=True, kw_only=True)
class MessageStatus(BaseUserUpdate):
    """
    Represents the status of a message.

    - `\`MessageStatus\` on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/components#statuses-object>`_.

    Attributes:
        id: The ID of the message that the status is for.
        metadata: The metadata of the message (to which phone number it was sent).
        status: The status of the message.
        timestamp: The timestamp when the status was updated.
        from_user: The user who the message was sent to.
        conversation: The conversation the given status notification belongs to (Optional).
        pricing_model: Type of pricing model used by the business. Current supported value is CBP.
        error: The error that occurred (if status is ``failed``).
    """
    id: str
    metadata: Metadata
    from_user: User
    timestamp: datetime
    status: MessageStatusType
    conversation: Conversation | None
    pricing_model: str
    error: WhatsAppError | None

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> MessageStatus:
        status = (value := update['entry'][0]['changes'][0]['value'])['statuses'][0]
        status_type = MessageStatusType(status['status'])
        return cls(
            _client=client,
            id=status['id'],
            metadata=Metadata.from_dict(value['metadata']),
            status=status_type,
            timestamp=datetime.fromtimestamp(int(status['timestamp'])),
            from_user=User(wa_id=status['recipient_id'], name=None),
            conversation=Conversation.from_dict(status['conversation']) if 'conversation' in status else None,
            pricing_model=status['pricing']['pricing_model'],
            error=WhatsAppError.from_incoming_error(status['errors'][0])
            if status_type == MessageStatusType.FAILED else None
        )


@dataclass(frozen=True, slots=True)
class Conversation:
    """
    Represents a conversation.

    - `\`Conversation\` on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/pricing#conversations>`_.

    Attributes:
        id: The ID of the conversation.
        category: The category of the conversation.
        expiration: The expiration date of the conversation (Optional, only for `sent` updates).

    """
    id: str
    category: ConversationCategory
    expiration: datetime | None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            id=data['id'],
            category=ConversationCategory(data['origin']['type']),
            expiration=datetime.fromtimestamp(int(data['expiration_timestamp']))
            if 'expiration_timestamp' in data else None
        )



from __future__ import annotations
import mimetypes
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from pywa import utils

if TYPE_CHECKING:
    from pywa.client import WhatsApp


@dataclass(frozen=True, slots=True, kw_only=True)
class MediaBase(ABC, utils.FromDict):
    """Base class for all media types."""

    _client: WhatsApp = field(repr=False, hash=False, compare=False)

    @property
    @abstractmethod
    def id(self) -> str: ...

    @property
    @abstractmethod
    def sha256(self) -> str: ...

    @property
    @abstractmethod
    def mime_type(self) -> str: ...

    def get_media_url(self) -> str:
        """Gets the URL of the media. (expires after 5 minutes)"""
        return self._client.get_media_url(media_id=self.id).url

    @property
    def extension(self) -> str | None:
        """Gets the extension of the media (with dot.)"""
        return mimetypes.guess_extension(self.mime_type)

    def download(
            self,
            path: str | None = None,
            filename: str | None = None,
            in_memory: bool = False,
    ) -> bytes | str:
        """
        Download a media file from WhatsApp servers.
            - Same as :func:`~pywa.client.WhatsApp.download_media` with ``media_url=media.get_media_url()``

        >>> message.image.download()

        Args:
            path: The path where to save the file (if not provided, the current working directory will be used).
            filename: The name of the file (if not provided, it will be guessed from the URL + extension).
            in_memory: Whether to return the file as bytes instead of saving it to disk (default: False).

        Returns:
            The path of the saved file if ``in_memory`` is False, the file as bytes otherwise.
        """
        return self._client.download_media(
            url=self.get_media_url(),
            path=path,
            filename=filename,
            in_memory=in_memory
        )


@dataclass(frozen=True, slots=True, kw_only=True)
class Image(MediaBase):
    """
    Represents an image.

    Attributes:
        id: The ID of the image.
        sha256: The SHA256 hash of the image.
        mime_type: The MIME type of the image.
    """
    id: str
    sha256: str
    mime_type: str


@dataclass(frozen=True, slots=True, kw_only=True)
class Video(MediaBase):
    """
    Represents a video.

    Attributes:
        id: The ID of the video.
        sha256: The SHA256 hash of the video.
        mime_type: The MIME type of the video.
    """
    id: str
    sha256: str
    mime_type: str


@dataclass(frozen=True, slots=True, kw_only=True)
class Sticker(MediaBase):
    """
    Represents a sticker.

    Attributes:
        id: The ID of the sticker.
        sha256: The SHA256 hash of the sticker.
        mime_type: The MIME type of the sticker.
        animated: Whether the sticker is animated.
    """
    id: str
    sha256: str
    mime_type: str
    animated: bool


@dataclass(frozen=True, slots=True, kw_only=True)
class Document(MediaBase):
    """
    Represents a document.

    Attributes:
        id: The ID of the document.
        sha256: The SHA256 hash of the document.
        mime_type: The MIME type of the document.
        filename: The filename of the document (optional).
    """
    id: str
    sha256: str
    mime_type: str
    filename: str | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class Audio(MediaBase):
    """
    Represents an audio.

    Attributes:
        id: The ID of the audio.
        sha256: The SHA256 hash of the audio.
        mime_type: The MIME type of the audio.
        voice: Whether the audio is a voice message or just an audio file.
    """
    id: str
    sha256: str
    mime_type: str
    voice: bool


@dataclass(frozen=True, slots=True)
class MediaUrlResponse(utils.FromDict):
    """
    Represents a media response.

    Attributes:
        id: The ID of the media.
        url: The URL of the media (valid for 5 minutes).
        mime_type: The MIME type of the media.
        sha256: The SHA256 hash of the media.
        file_size: The size of the media in bytes.
    """
    _client: WhatsApp = field(repr=False, hash=False, compare=False)
    id: str
    url: str
    mime_type: str
    sha256: str
    file_size: int

    def download(
            self,
            filepath: str | None = None,
            filename: str | None = None,
            in_memory: bool = False,
    ) -> bytes | str:
        """
        Download a media file from WhatsApp servers.

        Args:
            filepath: The path where to save the file (if not provided, the current working directory will be used).
            filename: The name of the file (if not provided, it will be guessed from the URL + extension).
            in_memory: Whether to return the file as bytes instead of saving it to disk (default: False).

        Returns:
            The path of the saved file if ``in_memory`` is False, the file as bytes otherwise.
        """
        return self._client.download_media(url=self.url, path=filepath, filename=filename, in_memory=in_memory)



from __future__ import annotations
from dataclasses import dataclass, field, asdict
from typing import Any, Iterable, TYPE_CHECKING
from pywa import utils

if TYPE_CHECKING:
    from pywa.client import WhatsApp


@dataclass(frozen=True, slots=True)
class User:
    """
    Represents a WhatsApp user.

    Attributes:
        wa_id: The WhatsApp ID of the user (The phone number with the country code).
        name: The name of the user (``None`` on MessageStatus).
    """
    wa_id: str
    name: str | None

    @classmethod
    def from_dict(cls, data: dict) -> User:
        return cls(wa_id=data["wa_id"], name=data["profile"]["name"])

    def as_vcard(self) -> str:
        """Get the user as a vCard."""
        return "\n".join((
            "BEGIN:VCARD",
            "VERSION:3.0",
            f"FN;CHARSET=UTF-8;ENCODING=QUOTED-PRINTABLE:{self.name}",
            f"TEL;type=CELL;type=VOICE:+{self.wa_id}",
            "END:VCARD"
        ))


class MessageType(utils.StrEnum):
    """
    Message types.

    Attributes:
        AUDIO: An audio message.
        DOCUMENT: A document message.
        TEXT: A text message.
        IMAGE: An image message.
        STICKER: A sticker message.
        VIDEO: A video message.
        REACTION: A reaction to a message.
        LOCATION: A location message.
        CONTACTS: A contacts message.
        INTERACTIVE: An interactive message (callback).
        ORDER: An order message.
        SYSTEM: A system message.
        UNKNOWN: An unknown message.
        UNSUPPORTED: An unsupported message.
    """
    AUDIO = "audio"
    DOCUMENT = "document"
    TEXT = "text"
    IMAGE = "image"
    STICKER = "sticker"
    VIDEO = "video"
    REACTION = "reaction"
    LOCATION = "location"
    CONTACTS = "contacts"
    INTERACTIVE = "interactive"
    ORDER = "order"
    SYSTEM = "system"
    UNKNOWN = "unknown"
    UNSUPPORTED = "unsupported"

    @classmethod
    def _missing_(cls, value: str) -> MessageType:
        return cls.UNSUPPORTED


@dataclass(frozen=True, slots=True)
class Reaction(utils.FromDict):
    """
    Represents a reaction to a message.

    Attributes:
        message_id: The ID of the message that was reacted to.
        emoji: The emoji that was used to react to the message (optional, ``None`` if removed).
    """
    message_id: str
    emoji: str | None = None


@dataclass(frozen=True, slots=True)
class Location(utils.FromDict):
    """
    Represents a location.

    Attributes:
        latitude: The latitude of the location.
        longitude: The longitude of the location.
        name: The name of the location (optional).
        address: The address of the location (optional).
        url: The URL of the location (optional).
    """
    latitude: float
    longitude: float
    name: str | None = None
    address: str | None = None
    url: str | None = None

    @property
    def current_location(self) -> bool:
        """Check if the shared location is the current location or manually selected."""
        return not any((self.name, self.address, self.url))

    def in_radius(self, lat: float, lon: float, radius: float | int) -> bool:
        """
        Check if the location is in a radius of another location.

        Args:
            lat: The latitude of the other location.
            lon: The longitude of the other location.
            radius: The radius in kilometers.
        """
        return utils.get_distance(lat1=self.latitude, lon1=self.longitude, lat2=lat, lon2=lon) <= radius


@dataclass(slots=True)
class Contact:
    """
    Represents a contact.

    Attributes:
        name: The name of the contact.
        birthday: The birthday of the contact (in ``YYYY-MM-DD`` format, optional).
        phones: The phone numbers of the contact.
        emails: The email addresses of the contact.
        urls: The URLs of the contact.
        addresses: The addresses of the contact.
        org: The organization of the contact (optional).
    """
    name: Name
    birthday: str | None = None
    phones: Iterable[Phone] = field(default_factory=tuple)
    emails: Iterable[Email] = field(default_factory=tuple)
    urls: Iterable[Url] = field(default_factory=tuple)
    addresses: Iterable[Address] = field(default_factory=tuple)
    org: Org | None = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            name=cls.Name.from_dict(data["name"]),
            birthday=data.get("birthday"),
            phones=tuple(cls.Phone.from_dict(phone) for phone in data.get("phones", ())),
            emails=tuple(cls.Email.from_dict(email) for email in data.get("emails", ())),
            urls=tuple(cls.Url.from_dict(url) for url in data.get("urls", ())),
            addresses=tuple(cls.Address.from_dict(address) for address in data.get("addresses", ())),
            org=cls.Org.from_dict(data["org"]) if "org" in data else None
        )

    def to_dict(self) -> dict[str, Any]:
        """Get the contact as a dict."""
        return {
            "name": asdict(self.name),
            "birthday": self.birthday,
            "phones": tuple(asdict(phone) for phone in self.phones),
            "emails": tuple(asdict(email) for email in self.emails),
            "urls": tuple(asdict(url) for url in self.urls),
            "addresses": tuple(asdict(address) for address in self.addresses),
            "org": asdict(self.org) if self.org else None,
        }

    def as_vcard(self) -> str:
        """Get the contact as a vCard."""
        return "\n".join(s for s in (
            "BEGIN:VCARD",
            "VERSION:3.0",
            f"FN;CHARSET=UTF-8;ENCODING=QUOTED-PRINTABLE:{self.name.formatted_name}",
            f"BDAY:{self.birthday}" if self.birthday else None,
            "\n".join(f"TEL;type={phone.type}:{phone.phone}" for phone in self.phones) if self.phones else None,
            "\n".join(f"EMAIL;type={email.type}:{email.email}" for email in self.emails) if self.emails else None,
            "\n".join(f"URL;type={url.type}:{url.url}" for url in self.urls) if self.urls else None,
            "\n".join(f"ADR;type={a.type}:;;{';'.join((getattr(a, f) or '') for f in ('street', 'city', 'state', 'zip', 'country') )}"
                      for a in self.addresses) if self.addresses else None,
            "END:VCARD"
        ) if s is not None)

    @dataclass(frozen=True, slots=True)
    class Name(utils.FromDict):
        """
        Represents a contact's name.

        - At least one of the optional parameters needs to be included along with the formatted_name parameter.

        Attributes:
            formatted_name: The formatted name of the contact.
            first_name: The first name of the contact (optional).
            last_name: The last name of the contact (optional).
            middle_name: The middle name of the contact (optional).
            suffix: The suffix of the contact (optional).
            prefix: The prefix of the contact (optional).
        """
        formatted_name: str
        first_name: str | None = None
        last_name: str | None = None
        middle_name: str | None = None
        suffix: str | None = None
        prefix: str | None = None

    @dataclass(frozen=True, slots=True)
    class Phone(utils.FromDict):
        """
        Represents a contact's phone number.

        Attributes:
            phone: The phone number (If ``wa_id`` is provided, No need for the ``phone``).
            type: The type of the phone number (Standard Values are CELL, MAIN, IPHONE, HOME, and WORK. optional).
            wa_id: The WhatsApp ID of the contact (optional).
        """
        phone: str | None = None
        type: str | None = None
        wa_id: str | None = None

    @dataclass(frozen=True, slots=True)
    class Email(utils.FromDict):
        """
        Represents a contact's email address.

        Attributes:
            email: The email address.
            type: The type of the email address (Standard Values are WORK and HOME. optional).
        """
        email: str | None = None
        type: str | None = None

    @dataclass(frozen=True, slots=True)
    class Url(utils.FromDict):
        """
        Represents a contact's URL.

        Attributes:
            url: The URL.
            type: The type of the URL (Standard Values are WORK and HOME. optional).
        """
        url: str | None = None
        type: str | None = None

    @dataclass(frozen=True, slots=True)
    class Org(utils.FromDict):
        """
        Represents a contact's organization.

        Attributes:
            company: The company of the contact (optional).
            department: The department of the contact (optional).
            title: The title of the business contact (optional).
        """
        company: str | None = None
        department: str | None = None
        title: str | None = None

    @dataclass(frozen=True, slots=True)
    class Address(utils.FromDict):
        """
        Represents a contact's address.

        Attributes:
            street: The street number and name of the address (optional).
            city: The city name of the address (optional).
            state: State abbreviation.
            zip: Zip code of the address (optional).
            country: Full country name.
            country_code: Two-letter country abbreviation (e.g. US, GB, IN. optional).
            type: The type of the address (Standard Values are WORK and HOME. optional).
        """
        street: str | None = None
        city: str | None = None
        state: str | None = None
        zip: str | None = None
        country: str | None = None
        country_code: str | None = None
        type: str | None = None


@dataclass(frozen=True, slots=True)
class ReplyToMessage:
    """
    Represents a message that was replied to.

    Attributes:
        message_id: The ID of the message that was replied to.
        from_user_id: The ID of the user who sent the message that was replied to.
    """
    message_id: str
    from_user_id: str

    @classmethod
    def from_dict(cls, data: dict | None) -> ReplyToMessage | None:
        return cls(
            message_id=data['id'],
            from_user_id=data['from']
        ) if (data and 'id' in data) else None


@dataclass(frozen=True, slots=True)
class Metadata(utils.FromDict):
    """
    Represents the metadata of a message.

    Attributes:
        display_phone_number: The phone number to which the message was sent.
        phone_number_id: The ID of the phone number to which the message was sent.
    """
    display_phone_number: str
    phone_number_id: str


@dataclass(frozen=True, slots=True)
class Product:
    """
    Represents a product in an order.

    Attributes:
        sku: Unique identifier of the product in a catalog (also referred to as ``Content ID`` or ``Retailer ID``).
        quantity: Number of items ordered.
        price: Price of the item.
        currency: Currency of the price.
    """
    sku: str
    quantity: int
    price: float
    currency: str

    @classmethod
    def from_dict(cls, data: dict) -> Product:
        return cls(
            sku=data["product_retailer_id"],
            quantity=data["quantity"],
            price=data["item_price"],
            currency=data["currency"]
        )

    @property
    def total_price(self) -> float:
        """Total price of the product."""
        return self.quantity * self.price


@dataclass(frozen=True, slots=True)
class Order:
    """
    Represents an order.

    Attributes:
        catalog_id: The ID for the catalog the ordered item belongs to.
        products:The ordered products.
        text: Text message from the user sent along with the order (optional).

    Properties:
        total_price: Total price of the order.
    """
    catalog_id: str
    products: tuple[Product]
    text: str | None

    @classmethod
    def from_dict(cls, data: dict, _client: WhatsApp) -> Order:
        return cls(
            catalog_id=data['catalog_id'],
            text=data.get('text'),
            products=tuple(Product.from_dict(p) for p in data['product_items'])
        )

    @property
    def total_price(self) -> float:
        """Total price of the order."""
        return sum(p.total_price for p in self.products)


@dataclass(frozen=True, slots=True)
class System:
    """
    Represents a system update (A customer has updated their phone number or profile information).

    Attributes:
        type: The type of the system update (``customer_changed_number`` or ``customer_identity_changed``).
        body: Describes the change to the customer's identity or phone number.
        identity: Hash for the identity fetched from server.
        wa_id: The WhatsApp ID for the customer prior to the update.
        new_wa_id: New WhatsApp ID for the customer when their phone number is updated.
    """
    type: str
    body: str
    identity: str
    wa_id: str
    new_wa_id: str | None

    @classmethod
    def from_dict(cls, data: dict, _client: WhatsApp) -> System:
        return cls(
            type=data['type'],
            body=data['body'],
            identity=data['identity'],
            wa_id=data['customer'],
            new_wa_id=data.get('wa_id')
        )


@dataclass(frozen=True, slots=True)
class ProductsSection:
    """
    Represents a section in a section list.

    Attributes:
        title: The title of the products section (up to 24 characters).
        skus: The SKUs of the products in the section (at least 1, no more than 30).
    """
    title: str
    skus: Iterable[str]

    def to_dict(self) -> dict:
        return {
            "title": self.title,
            "product_items": tuple({"product_retailer_id": sku} for sku in self.skus)
        }


class Industry(utils.StrEnum):
    """
    Represents the industry of a business.

    Attributes:
        UNDEFINED: Undefined.
        OTHER: Other.
        AUTO: Automotive.
        BEAUTY: Beauty.
        APPAREL: Apparel.
        EDU: Education.
        ENTERTAIN: Entertainment.
        EVENT_PLAN: Event planning.
        FINANCE: Finance.
        GROCERY: Grocery store.
        GOVT: Government.
        HOTEL: Hotel.
        HEALTH: Health.
        NONPROFIT: Nonprofit.
        PROF_SERVICES: Professional services.
        RETAIL: Retail.
        TRAVEL: Travel.
        RESTAURANT: Restaurant.
        NOT_A_BIZ: Not a business.
    """
    UNDEFINED = "UNDEFINED"
    OTHER = "OTHER"
    AUTO = "AUTO"
    BEAUTY = "BEAUTY"
    APPAREL = "APPAREL"
    EDU = "EDU"
    ENTERTAIN = "ENTERTAIN"
    EVENT_PLAN = "EVENT_PLAN"
    FINANCE = "FINANCE"
    GROCERY = "GROCERY"
    GOVT = "GOVT"
    HOTEL = "HOTEL"
    HEALTH = "HEALTH"
    NONPROFIT = "NONPROFIT"
    PROF_SERVICES = "PROF_SERVICES"
    RETAIL = "RETAIL"
    TRAVEL = "TRAVEL"
    RESTAURANT = "RESTAURANT"
    NOT_A_BIZ = "NOT_A_BIZ"

    @classmethod
    def _missing_(cls, value: str) -> None:
        return None


@dataclass(frozen=True, slots=True)
class BusinessProfile:
    """
    Represents a business profile.

    Attributes:
        about: This text appears in the business's profile, beneath its profile image, phone number, and contact buttons.
        address: Address of the business. Character limit 256.
        description: Description of the business. Character limit 512.
        email: The contact email address (in valid email format) of the business. Character limit 128.
        industry: The industry of the business.
        profile_picture_url: URL of the profile picture that was uploaded to Meta.
        websites: The URLs associated with the business. For instance, a website, Facebook Page, or Instagram.
         There is a maximum of 2 websites with a maximum of 256 characters each.
    """
    about: str
    address: str | None
    industry: Industry
    description: str | None
    email: str | None
    profile_picture_url: str | None
    websites: tuple[str] | None

    @classmethod
    def from_dict(cls, data: dict) -> BusinessProfile:
        return cls(
            about=data['about'],
            address=data.get('address'),
            industry=Industry(data['vertical']),
            description=data.get('description'),
            email=data.get('email'),
            profile_picture_url=data.get('profile_picture_url'),
            websites=tuple(data.get('websites', [])) or None
        )


@dataclass(frozen=True, slots=True)
class CommerceSettings:
    """
    Represents the WhatsApp commerce settings.

    Attributes:
        catalog_id: The ID of the catalog associated with the business.
        is_catalog_visible: Whether the catalog is visible to customers.
        is_cart_enabled: Whether the cart is enabled.
    """
    catalog_id: str
    is_catalog_visible: bool
    is_cart_enabled: bool

    @classmethod
    def from_dict(cls, data: dict) -> CommerceSettings:
        return cls(
            catalog_id=data['id'],
            is_catalog_visible=data['is_catalog_visible'],
            is_cart_enabled=data['is_cart_enabled']
        )




"""
This module contains the handlers for incoming updates.
"""

from __future__ import annotations

__all__ = [
    "MessageHandler",
    "CallbackButtonHandler",
    "CallbackSelectionHandler",
    "RawUpdateHandler",
    "MessageStatusHandler"
]

import dataclasses
import abc
from typing import Callable, Any, TYPE_CHECKING, Iterable, cast
from pywa import filters as fil
from pywa.types import Message, CallbackButton, CallbackSelection, MessageStatus, TemplateStatus
from pywa.types.base_update import BaseUpdate
from pywa.types.callback import CallbackDataT, CallbackData

if TYPE_CHECKING:
    from pywa.client import WhatsApp


def _resolve_callback_data(factory: CallbackDataT) -> tuple[CallbackDataT, tuple[Callable[[WhatsApp, Any], bool]]]:
    """Internal function to resolve the callback data into a constractor and a filter."""
    clb_filter = tuple()
    if isinstance(factory, Iterable):
        constractor = lambda data: ( # noqa
            tuple(map(lambda fs: (fs[0].from_str if issubclass(fs[0], CallbackData) else fs[0])(fs[1]), # noqa
                      zip(factory, data.split(CallbackData.__callback_sep__))))
        )
        if any((callback_datas := tuple(bool(issubclass(f, CallbackData)) for f in factory))):
            def _clb_filter(_: WhatsApp, btn_or_sel: CallbackButton | CallbackSelection) -> bool:
                data = btn_or_sel.data.split(CallbackData.__callback_sep__)
                return all(
                    data[i].startswith(
                        str(cast(CallbackData, factory[i]).__callback_id__) +
                        cast(CallbackData, factory[i]).__callback_data_sep__
                    )
                    for i, b in enumerate(callback_datas) if b
                )
            clb_filter = (_clb_filter,)

    elif issubclass(factory, CallbackData):
        constractor = factory.from_str
        clb_filter = (fil.callback.data_startswith(str(factory.__callback_id__) + factory.__callback_data_sep__),)
    elif callable(factory):
        constractor = factory
    else:
        raise ValueError(f"Unsupported factory type {factory}.")
    return constractor, clb_filter


class Handler(abc.ABC):
    """Base class for all handlers."""
    __field_name__: str
    """The field name of the webhook update
    https://developers.facebook.com/docs/graph-api/webhooks/reference/whatsapp-business-account"""

    __constructor__: Callable[[WhatsApp, dict], BaseUpdate]
    """The constructor to use to construct the update object from the webhook update dict."""

    def __init__(
            self,
            handler: Callable[[WhatsApp, Any], Any],
            *filters: Callable[[WhatsApp, Any], bool]
    ):
        """
        Initialize a new handler.
        """
        self.handler = handler
        self.filters = filters

    def __call__(self, wa: WhatsApp, data: Any):
        if all((f(wa, data) for f in self.filters)):
            self.handler(wa, data)

    @classmethod
    def __hash__(cls) -> int:
        return hash(cls.__field_name__)


class MessageHandler(Handler):
    """
    Handler for incoming :class:`pywa.types.Message`.

    - You can use the :func:`~pywa.client.WhatsApp.on_message` decorator to register a handler for this type.

    Example:

        >>> from pywa import WhatsApp, filters as fil
        >>> wa = WhatsApp(...)
        >>> print_text_messages = lambda _, msg: print(msg)
        >>> wa.add_handlers(MessageHandler(print_text_messages, fil.text))

    Args:
        handler: The handler function (gets the :class:`pywa.WhatsApp` instance and a :class:`pywa.types.Message` as
            arguments)
        *filters: The filters to apply to the handler (gets a :class:`pywa.WhatsApp` instance and a
            :class:`pywa.types.Message` and returns a :class:`bool`)
    """
    __field_name__ = "messages:msg"
    __constructor__ = Message.from_update

    def __init__(
            self,
            handler: Callable[[WhatsApp, Message], Any],
            *filters: Callable[[WhatsApp, Message], bool]
    ):
        super().__init__(handler, *filters)


class CallbackButtonHandler(Handler):
    """
    Handler for callback buttons (User clicks on a :class:`pywa.types.Button`).

    - You can use the :func:`~pywa.client.WhatsApp.on_callback_button` decorator to register a handler for this type.

    **IMPORTANT:** If ``factory`` provided, The filters have no access to the constructed callback data, only the raw
    string (When subclassing :class:`pywa.types.CallbackData` as the factory, a matching filter is automatically
    added, so there is no need to add it manually).

    Example:

        >>> from pywa import WhatsApp, filters as fil
        >>> wa = WhatsApp(...)
        >>> print_btn = lambda _, btn: print(btn)
        >>> wa.add_handlers(CallbackButtonHandler(print_btn, fil.callback.data_startswith('id:')))

    Args:
        handler: The handler function (gets the WhatsApp instance and the callback as arguments)
        *filters: The filters to apply to the handler (gets a :class:`pywa.WhatsApp` instance and a
            :class:`pywa.types.CallbackButton` and returns a :class:`bool`)
        factory: The constructor/s to use to construct the callback data (default: :class:`str`).
    """
    __field_name__ = "messages:btn"
    __constructor__ = CallbackButton.from_update

    def __init__(
            self,
            handler: Callable[[WhatsApp, CallbackButton], Any],
            *filters: Callable[[WhatsApp, CallbackButton], bool],
            factory: CallbackDataT = str
    ):
        self.factory, clb_filter = _resolve_callback_data(factory)
        super().__init__(handler, *clb_filter, *filters)

    def __call__(self, wa: WhatsApp, data: CallbackButton):
        if all((f(wa, data) for f in self.filters)):
            self.handler(
                wa,
                dataclasses.replace(data, data=self.factory(data.data))
                if not issubclass(self.factory, str) else data
            )


class CallbackSelectionHandler(Handler):
    """
    Handler for callback selections (User selects an option from :class:`pywa.types.SectionList`).

    - You can use the :func:`~pywa.client.WhatsApp.on_callback_selection` decorator to register a handler for this type.

    **IMPORTANT:** If ``factory`` provided, The filters have no access to the constructed callback data, only the raw
    string (When subclassing :class:`pywa.types.CallbackData` as the factory, a matching filter is automatically
    added, so there is no need to add it manually).

    Example:

        >>> from pywa import WhatsApp, filters as fil
        >>> wa = WhatsApp(...)
        >>> print_selection = lambda _, sel: print(sel)
        >>> wa.add_handlers(CallbackSelectionHandler(print_selection, fil.callback.data_startswith('id:')))

    Args:
        handler: The handler function. (gets a :class:`pywa.WhatsApp` instance and a
            :class:`pywa.types.CallbackSelection` as arguments)
        *filters: The filters to apply to the handler. (gets a :class:`pywa.WhatsApp` instance and a
            :class:`pywa.types.CallbackSelection` and returns a :class:`bool`)
        factory: The constructor/s to use to construct the callback data (default: :class:`str`).
    """
    __field_name__ = "messages:sel"
    __constructor__ = CallbackSelection.from_update

    def __init__(
            self,
            handler: Callable[[WhatsApp, CallbackSelection], Any],
            *filters: Callable[[WhatsApp, CallbackSelection], bool],
            factory: CallbackDataT = str
    ):
        self.factory, clb_filter = _resolve_callback_data(factory)
        super().__init__(handler, *clb_filter, *filters)

    def __call__(self, wa: WhatsApp, data: CallbackSelection):
        if all((f(wa, data) for f in self.filters)):
            self.handler(
                wa,
                dataclasses.replace(data, data=self.factory(data.data))
                if not issubclass(self.factory, str) else data
            )


class MessageStatusHandler(Handler):
    """
    Handler for :class:`pywa.types.MessageStatus` updates (Message is sent, delivered, read, failed, etc...).

    - You can use the :func:`~pywa.client.WhatsApp.on_message_status` decorator to register a handler for this type.

    **DO NOT USE THIS HANDLER WITHOUT FILTERS TO SEND MESSAGES, IT WILL CAUSE AN INFINITE LOOP!**

    Example:

        >>> from pywa import WhatsApp, filters as fil
        >>> wa = WhatsApp(...)
        >>> print_failed_messages = lambda _, msg: print(msg)
        >>> wa.add_handlers(MessageStatusHandler(print_failed_messages, fil.message_status.failed))

    Args:
        handler: The handler function (gets a :class:`pywa.WhatsApp` instance and a :class:`pywa.types.MessageStatus` as
            arguments)
        *filters: The filters to apply to the handler (gets a :class:`pywa.WhatsApp` instance and a
            :class:`pywa.types.MessageStatus` and returns a :class:`bool`)
    """
    __field_name__ = "messages:status"
    __constructor__ = MessageStatus.from_update

    def __init__(
            self,
            handler: Callable[[WhatsApp, MessageStatus], Any],
            *filters: Callable[[WhatsApp, MessageStatus], bool]
    ):
        super().__init__(handler, *filters)


class TemplateStatusHandler(Handler):
    """
    Handler for :class:`pywa.types.TemplateStatus` updates (Template message is approved, rejected etc...).

    - You can use the :func:`~pywa.client.WhatsApp.on_template_status` decorator to register a handler for this type.


    Example:

        >>> from pywa import WhatsApp, filters as fil
        >>> wa = WhatsApp(...)
        >>> print_template_status = lambda _, msg: print(msg)
        >>> wa.add_handlers(TemplateStatusHandler(
        ...     print_template_status,
        ...     fil.template_status.on_event(TemplateStatus.TemplateEvent.APPROVED)
        ... ))

    Args:
        handler: The handler function (gets a :class:`pywa.WhatsApp` instance and a
            :class:`pywa.types.TemplateStatus` as arguments)
        *filters: The filters to apply to the handler (gets a :class:`pywa.WhatsApp` instance and a
            :class:`pywa.types.TemplateStatus` and returns a :class:`bool`)
    """
    __field_name__ = "message_template_status_update"
    __constructor__ = TemplateStatus.from_update

    def __init__(
            self,
            handler: Callable[[WhatsApp, TemplateStatus], Any],
            *filters: Callable[[WhatsApp, TemplateStatus], bool]
    ):
        super().__init__(handler, *filters)


class RawUpdateHandler(Handler):
    """
    A raw update handler.

    - This handler will be called for **EVERY** update received from WhatsApp, even if it's not sent to the client phone number.
    - You can use the :func:`~pywa.client.WhatsApp.on_raw_update` decorator to register a handler for this type.

    Example:

        >>> from pywa import WhatsApp
        >>> wa = WhatsApp(...)
        >>> print_updates = lambda _, data: print(data)
        >>> wa.add_handlers(RawUpdateHandler(print_updates))

    Args:
        handler: The handler function (gets a :class:`pywa.WhatsApp` instance and a :class:`dict` as arguments)
        *filters: The filters to apply to the handler (gets a :class:`pywa.WhatsApp` instance and a :class:`dict` and
            returns a :class:`bool`)
    """
    __field_name__ = "raw"
    __constructor__ = lambda _, data: data  # noqa

    def __init__(
            self,
            handler: Callable[[WhatsApp, dict], Any],
            *filters: Callable[[WhatsApp, dict], bool]
    ):
        super().__init__(handler, *filters)


class HandlerDecorators:
    def __init__(self):
        raise TypeError("This class cannot be instantiated.")

    def on_raw_update(
            self: WhatsApp, *filters: Callable[[WhatsApp, dict], bool]
    ) -> Callable[[Callable[[WhatsApp, dict], Any]], Callable[[WhatsApp, dict], Any]]:
        """
        Decorator to register a function as a handler for raw updates (:class:`dict`).

        - This handler is called for **EVERY** update received from WhatsApp, even if it's not sent to the client phone number.
        - Shortcut for :func:`~pywa.client.WhatsApp.add_handlers` with a :class:`RawUpdateHandler`.

        Example:

            >>> wa = WhatsApp(...)
            >>> @wa.on_raw_update()
            ... def raw_update_handler(_: WhatsApp, update: dict):
            ...     print(update)

        Args:
            *filters: Filters to apply to the incoming updates (filters are function that take a :class:`pywa.WhatsApp`
                instance and the incoming update :class:`dict` and return a :class:`bool` if the update should be
                handled).
        """
        def decorator(func: Callable[[WhatsApp, dict], Any]) -> Callable[[WhatsApp, dict], Any]:
            self.add_handlers(RawUpdateHandler(func, *filters))
            return func
        return decorator

    def on_message(
            self: WhatsApp, *filters: Callable[[WhatsApp, Message], bool]
    ) -> Callable[[Callable[[WhatsApp, Message], Any]], Callable[[WhatsApp, Message], Any]]:
        """
        Decorator to register a function as a handler for incoming :class:`pywa.types.Message` (User sends a message).

        - Shortcut for :func:`~pywa.client.WhatsApp.add_handlers` with a :class:`MessageHandler`.

        Example:

            >>> from pywa.types import Button
            >>> from pywa import filters as fil
            >>> wa = WhatsApp(...)
            >>> @wa.on_message(fil.text.matches("Hello", "Hi", ignore_case=True))
            ... def hello_handler(_: WhatsApp, msg: Message):
            ...     msg.react("👋")
            ...     msg.reply_text(text="Hello from PyWa!", quote=True, buttons=[Button("Help", data="help")

        Args:
            *filters: Filters to apply to the incoming messages (filters are function that take a :class:`pywa.WhatsApp`
                instance and the incoming :class:`pywa.types.Message` and return a boolean).
        """
        def decorator(func: Callable[[WhatsApp, Message], Any]) -> Callable[[WhatsApp, Message], Any]:
            self.add_handlers(MessageHandler(func, *filters))
            return func
        return decorator

    def on_callback_button(
            self: WhatsApp, *filters: Callable[[WhatsApp, CallbackButton], bool], factory: CallbackDataT = str
    ) -> Callable[[Callable[[WhatsApp, CallbackButton], Any]], Callable[[WhatsApp, CallbackButton], Any]]:
        """
        Decorator to register a function as a handler when a user clicks on a :class:`pywa.types.Button`.

        - Shortcut for :func:`~pywa.client.WhatsApp.add_handlers` with a :class:`CallbackButtonHandler`.

        **IMPORTANT:** If ``factory`` provided, The filters have no access to the constructed callback data, only the raw
        string (When subclassing :class:`pywa.types.CallbackData` as the factory, a matching filter is automatically
        added, so there is no need to add it manually).

        Example:

            >>> from pywa.types import CallbackButton
            >>> from pywa import filters as fil
            >>> wa = WhatsApp(...)
            >>> @wa.on_callback_button(fil.callback.data_matches("help"))
            ... def help_handler(_: WhatsApp, btn: CallbackButton):
            ...     btn.reply_text(text="What can I help you with?")

        Args:
            *filters: Filters to apply to the incoming callback button presses (filters are function that take a
             :class:`pywa.WhatsApp` instance and the incoming :class:`pywa.types.CallbackButton` and return :class:`bool`).
            factory: The constructor/s to use for the callback data (default: :class:`str`).
        """
        def decorator(func: Callable[[WhatsApp, CallbackButton], Any]) -> Callable[[WhatsApp, CallbackButton], Any]:
            self.add_handlers(CallbackButtonHandler(func, *filters, factory=factory))
            return func
        return decorator

    def on_callback_selection(
            self: WhatsApp, *filters: Callable[[WhatsApp, CallbackSelection], bool], factory: CallbackDataT = str
    ) -> Callable[[Callable[[WhatsApp, CallbackSelection], Any]], Callable[[WhatsApp, CallbackSelection], Any]]:
        """
        Decorator to register a function as a handler when a user selects an option from a :class:`pywa.types.SectionList`.

        - Shortcut for :func:`~pywa.client.WhatsApp.add_handlers` with a :class:`CallbackSelectionHandler`.

        **IMPORTANT:** If ``factory`` provided, The filters have no access to the constructed callback data, only the raw
        string (When subclassing :class:`pywa.types.CallbackData` as the factory, a matching filter is automatically
        added, so there is no need to add it manually).

        Example:

            >>> from pywa.types import CallbackSelection
            >>> from pywa import filters as fil
            >>> wa = WhatsApp(...)
            >>> @wa.on_callback_selection(fil.callback.data_startswith("id:"))
            ... def id_handler(_: WhatsApp, sel: CallbackSelection):
            ...     sel.reply_text(text=f"Your ID is {sel.data.split(':', 1)[1]}")

        Args:
            *filters: Filters to apply to the incoming callback selections (filters are function that take a
                :class:`pywa.WhatsApp` instance and the incoming :class:`pywa.types.CallbackSelection` and return :class:`bool`).
            factory: The constructor/s to use for the callback data (default: :class:`str`).
        """
        def decorator(
                func: Callable[[WhatsApp, CallbackSelection], Any]
        ) -> Callable[[WhatsApp, CallbackSelection], Any]:
            self.add_handlers(CallbackSelectionHandler(func, *filters, factory=factory))
            return func
        return decorator

    def on_message_status(
            self: WhatsApp, *filters: Callable[[WhatsApp, MessageStatus], bool]
    ) -> Callable[[Callable[[WhatsApp, MessageStatus], Any]], Callable[[WhatsApp, MessageStatus], Any]]:
        """
        Decorator to register a function as a handler for incoming message status changes (Message is sent, delivered,
        read, failed, etc...).

        - Shortcut for :func:`~pywa.client.WhatsApp.add_handlers` with a :class:`MessageStatusHandler`.

        **DO NOT USE THIS HANDLER WITHOUT FILTERS TO SEND MESSAGES, IT WILL CAUSE AN INFINITE LOOP!**

        Example:

            >>> from pywa.types import MessageStatus
            >>> from pywa import filters as fil
            >>> wa = WhatsApp(...)
            >>> @wa.on_message_status(fil.message_status.failed)
            ... def delivered_handler(client: WhatsApp, status: MessageStatus):
            ...     print(f"Message {status.id} failed to send to {status.from_user.wa_id}: {status.error.message})


        Args:
        *filters: Filters to apply to the incoming message status changes (filters are function that take a
            :class:`pywa.WhatsApp` instance and the incoming :class:`pywa.types.MessageStatus` and return :class:`bool`).
        """
        def decorator(func: Callable[[WhatsApp, MessageStatus], Any]) -> Callable[[WhatsApp, MessageStatus], Any]:
            self.add_handlers(MessageStatusHandler(func, *filters))
            return func
        return decorator

    def on_template_status(
            self: WhatsApp, *filters: Callable[[WhatsApp, TemplateStatus], bool]
    ) -> Callable[[Callable[[WhatsApp, TemplateStatus], Any]], Callable[[WhatsApp, TemplateStatus], Any]]:
        """
        Decorator to register a function as a handler for :class:`pywa.types.TemplateStatus` updates (Template message
        is approved, rejected etc...).

        - Shortcut for :func:`~pywa.client.WhatsApp.add_handlers` with a :class:`TemplateStatusHandler`.

        Example:

            >>> from pywa.types import TemplateStatus
            >>> from pywa import filters as fil
            >>> wa = WhatsApp(...)
            >>> @wa.on_template_status(fil.template_status.on_event(TemplateStatus.TemplateEvent.APPROVED))
            ... def approved_handler(client: WhatsApp, status: TemplateStatus):
            ...     print(f"Template {status.message_template_name} just got approved!")

        Args:
            *filters: Filters to apply to the incoming template status changes (filters are function that take a
                :class:`pywa.WhatsApp` instance and the incoming :class:`pywa.types.TemplateStatus` and return :class:`bool`).
        """
        def decorator(func: Callable[[WhatsApp, TemplateStatus], Any]) -> Callable[[WhatsApp, TemplateStatus], Any]:
            self.add_handlers(TemplateStatusHandler(func, *filters))
            return func
        return decorator

